from iktomi.db.sqla.images import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.db.sqla.images is deprecated. Use ' # pragma: no cover
              'iktomi.db.sqla.images instead',
              category=DeprecationWarning,
              stacklevel=2)



from iktomi.forms.files import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.forms.files is deprecated. Use ' # pragma: no cover
              'iktomi.forms.files instead',
              category=DeprecationWarning)



# -*- coding: utf-8 -*-
import six

from xml.sax import saxutils
import weakref, re, sys

from iktomi.utils.i18n import M_, N_ # deprecated, import from iktomi.utils.i18n

def quoteattr(value):
    '''Works like quoteattr from saxutils (returns escaped string in quotes),
    but is safe for HTML'''
    if value == '':
        return '""'
    return '"{}"'.format(saxutils.escape(value, {'"': '&quot;'}))

def quoteattrs(data):
    '''Takes dict of attributes and returns their HTML representation'''
    items = []
    for key, value in data.items():
        items.append('{}={}'.format(key, quoteattr(value)))
    return ' '.join(items)

def quote_js(text):
    '''Quotes text to be used as JavaScript string in HTML templates. The
    result doesn't contain surrounding quotes.'''
    if isinstance(text, six.binary_type):
        text = text.decode('utf-8') # for Jinja2 Markup
    text = text.replace('\\', '\\\\');
    text = text.replace('\n', '\\n');
    text = text.replace('\r', '');
    for char in '\'"<>&':
        text = text.replace(char, '\\x{:02x}'.format(ord(char)))
    return text

def weakproxy(obj):
    '''Safe version of weakref.proxy.'''
    try:
        obj = weakref.proxy(obj)
    except TypeError:
        pass
    return obj


class cached_property(object):
    '''Turns decorated method into caching property (method is called once on
    first access to property).'''

    def __init__(self, method, name=None):
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__

    def __get__(self, inst, cls):
        if inst is None:
            return self
        result = self.method(inst)
        setattr(inst, self.name, result)
        return result


class cached_class_property(object):
    '''Turns decorated method into caching class property (method is called
    once on first access to property of class or any of its instances).'''

    def __init__(self, method, name=None):
        self.method = method
        self.name = name or method.__name__

    def __get__(self, inst, cls):
        result = self.method(cls)
        setattr(cls, self.name, result)
        return result


# http://www.w3.org/TR/REC-xml/#NT-Char
# Char ::= #x9 | #xA | #xD | [#x20-#xD7FF] | [#xE000-#xFFFD] | 
#          [#x10000- #x10FFFF]
# (any Unicode character, excluding the surrogate blocks, FFFE, and FFFF)
_char_tail = ''
if sys.maxunicode > 0x10000:
    _char_tail = u'{}-{}'.format(six.unichr(0x10000),
                                 six.unichr(min(sys.maxunicode, 0x10FFFF)))
_nontext_sub = re.compile(
            u'[^\\x09\\x0A\\x0D\\x20-\uD7FF\uE000-\uFFFD{}]'.format(_char_tail),
            re.U).sub
def replace_nontext(text, replacement=u'\uFFFD'):
    return _nontext_sub(replacement, text)





# -*- coding: utf-8 -*-

import os
import sys
import time
import errno
import signal
import logging
logger = logging.getLogger(__name__)

from .base import Cli
from iktomi.utils.system import safe_makedirs, is_running, terminate, \
                                doublefork


def flup_fastcgi(wsgi_app, bind, cwd=None, pidfile=None, logfile=None,
                 daemonize=False, umask=None, **params):
    if params.pop('preforked', False):
        from flup.server import fcgi_fork as fcgi
    else:
        from flup.server import fcgi
    if daemonize:
        if os.path.isfile(pidfile):
            with open(pidfile, 'r') as f:
                try:
                    pid = int(f.read())
                except ValueError:
                    pid = None

            if pid is not None and  is_running(pid):
                sys.exit('Already running (PID: {})'.format(pid))
            elif pid is not None:
                logger.info('PID file was pointing to nonexistent process %r',
                            pid)
            else:
                logger.info('PID file should contain a number')
        doublefork(pidfile, logfile, cwd, umask)
    logger.info('Starting FastCGI server (flup), current working dir %r', cwd)
    fcgi.WSGIServer(wsgi_app, bindAddress=bind, umask=umask,
                    debug=False, **params).run()


class Flup(Cli):
    '''
    Flup FastCGI server

    :param app: iktomi app
    :param bind: socket file
    :param logfile: log file
    :param pidfile: PID file
    :param cwd: current working directory
    :param umask:
    :param dict fastcgi_params: arguments accepted by flup `WSGIServer`,
        plus `preforked`
    '''

    def __init__(self, app, bind='', logfile=None, pidfile=None,
                 cwd='.', umask=2, fastcgi_params=None):
        self.app = app
        self.cwd = os.path.abspath(cwd)
        if ':' in bind:
            host, port = bind.split(':')
            port = int(port)
            bind = (host, port)
        else:
            bind = os.path.abspath(bind or os.path.join(self.cwd, 'fcgi.sock'))
            safe_makedirs(bind)
        self.bind = bind
        self.umask = umask
        self.logfile = logfile or os.path.join(self.cwd, 'fcgi.log')
        self.pidfile = pidfile or os.path.join(self.cwd, 'fcgi.pid')
        self.fastcgi_params = fastcgi_params or {}

    def command_start(self, daemonize=False):
        '''
        Start a server::

            ./manage.py flup:start [--daemonize]
        '''
        if daemonize:
            safe_makedirs(self.logfile, self.pidfile)
        flup_fastcgi(self.app, bind=self.bind, pidfile=self.pidfile,
                     logfile=self.logfile, daemonize=daemonize,
                     cwd=self.cwd, umask=self.umask, **self.fastcgi_params)

    def command_stop(self):
        '''
        Stop a server::

            ./manage.py flup:stop
        '''
        if self.pidfile:
            if not os.path.exists(self.pidfile):
                sys.exit("Pidfile {!r} doesn't exist".format(self.pidfile))
            with open(self.pidfile) as pidfile:
                pid = int(pidfile.read())
            for sig in [signal.SIGINT, signal.SIGTERM, signal.SIGKILL]:
                try:
                    if terminate(pid, sig, 3):
                        os.remove(self.pidfile)
                        # NOTE: we are not performing sys.exit here,
                        # otherwise restart command will not work
                        return
                except OSError as exc:
                    if exc.errno != errno.ESRCH:
                        raise
                    elif sig == signal.SIGINT:
                        sys.exit('Not running')
        sys.exit('No pidfile provided')

    def command_restart(self):
        self.command_stop()
        # restart is useless for non-daemon programs
        self.command_start(daemonize=True)



# -*- coding: utf-8 -*-
from six.moves.urllib.parse import urlsplit
from lxml import html
from lxml.html import clean
# XXX move to iktomi.cms?

class Cleaner(clean.Cleaner):

    safe_attrs_only = True
    remove_unknown_tags = None
    drop_empty_tags = frozenset()
    dom_callbacks = []
    allow_external_src = False
    allowed_protocols = frozenset(['http', 'https', 'mailto'])
    # None to allow all classes
    allow_classes = None
    attr_val_is_uri = ['href', 'src', 'cite', 'action', 'longdesc']
    a_without_href = True
    # False : no tags wrapping;
    # None : try to wrap tags on top in 'p' if 'p' is allowed or 'div'
    # True : try to wrap tags on top in 'p' if 'p' is allowed or 'div',
    #    and raise error if no get_wrapper_tag was found
    # if div allowed;
    # 'div'/'p' : wrap tags in 'div' or 'p' respectively
    # lambda : wrap tags in tag from lambda
    wrap_inline_tags = None
    # Tags to wrap in paragraphs on top
    tags_to_wrap = ['b', 'big', 'i', 'small', 'tt',
                    'abbr', 'acronym', 'cite', 'code',
                    'dfn', 'em', 'kbd', 'strong', 'samp',
                    'var', 'a', 'bdo', 'br', 'map', 'object',
                    'q', 'span', 'sub', 'sup']
    split_paragraphs_by_br = True

    def __init__(self, *args, **kwargs):
        clean.Cleaner.__init__(self, *args, **kwargs)
        if self.wrap_inline_tags is True:
            if self.get_wrapper_tag() is None:
                raise ValueError('Cannot find top element')

    def __call__(self, doc):
        clean.Cleaner.__call__(self, doc)
        if hasattr(doc, 'getroot'):
            # ElementTree instance, instead of an element
            doc = doc.getroot()
        self.extra_clean(doc)

    # retrieve tag to wrap around inline tags
    def get_wrapper_tag(self):
        if self.allow_tags is None:
            return
        if self.wrap_inline_tags in (None, True):
            if 'p' in self.allow_tags:
                return html.Element('p')
            elif 'div' in self.allow_tags:
                return html.Element('div')
        elif self.wrap_inline_tags in ('p', 'div'):
            if 'p' in self.allow_tags or 'div' in self.allow_tags:
                return html.Element(self.wrap_inline_tags)
        elif callable(self.wrap_inline_tags):
            element = self.wrap_inline_tags()
            if element.tag in self.allow_tags:
                return element

    def clean_top(self, doc):
        par = None
        first_par = False
        if self.get_wrapper_tag() is None:
            return
        # create paragraph if there text in the beginning of top
        if (doc.text or "").strip():
            par = self.get_wrapper_tag()
            doc.insert(0, par)
            par.text = doc.text
            doc.text = None
            # remember if first paragraph created from text
            first_par = True

        for child in doc.getchildren():
            i = doc.index(child)

            if child.tag == 'br' and 'br' in self.tags_to_wrap:
                if (child.tail or "").strip():
                    par = self.get_wrapper_tag()
                    doc.insert(i, par)
                    par.text = child.tail
                doc.remove(child)
                continue

            if child.tag not in self.tags_to_wrap and \
                    (child.tail or "").strip():
                par = self.get_wrapper_tag()
                par.text = child.tail
                child.tail = None
                doc.insert(i+1, par)
                continue

            if child.tag in self.tags_to_wrap:
                if par is None:
                    par = self.get_wrapper_tag()
                    doc.insert(i, par)
                par.append(child)
            else:
                if first_par and i == 0:
                    continue
                par = None

    def _tail_is_empty(self, el):
        return not (el.tail and el.tail.strip(u'  \t\r\n\v\f\u00a0'))

    def is_element_empty(self, el):
        if el.tag == 'br':
            return True
        if el.tag not in self.drop_empty_tags:
            return False
        children = el.getchildren()
        empty_children = all(
            [self.is_element_empty(child) and self._tail_is_empty(child)
             for child in children]
        )
        text = el.text and el.text.strip(u'  \t\r\n\v\f\u00a0')
        return not text and empty_children

    def remove_brs_from_pars(self, doc):
        def split_by_br(par):
            br = par.find('.//br')
            if br is not None:
                next_par = html.Element('p')
                par.addnext(next_par)
                if br.tail:
                    next_par.text = br.tail
                    br.tail = None
                nxt = br.getnext()
                while nxt is not None:
                    next_par.append(nxt)
                    nxt = br.getnext()
                br.drop_tag()
                split_by_br(next_par)
        for p in doc.findall('.//p'):
            split_by_br(p)

    def extra_clean(self, doc):
        for el in doc.xpath('//*[@href]'):
            scheme, netloc, path, query, fragment = urlsplit(el.attrib['href'])
            if scheme and scheme not in self.allowed_protocols:
                el.drop_tag()

        for attr in self.attr_val_is_uri:
            if attr == 'href':
                continue
            for el in doc.xpath('//*[@'+attr+']'):
                scheme, netloc, path, query, fragment = urlsplit(el.attrib[attr])
                scheme_fail = scheme and scheme not in self.allowed_protocols
                netloc_fail = not self.allow_external_src and netloc
                if scheme_fail or netloc_fail:
                    if attr == 'src':
                        el.drop_tag()
                    else:
                        el.attrib.pop(attr)

        if self.a_without_href:
            for link in doc.xpath('//a[not(@href)]'):
                link.drop_tag()

        if self.allow_classes is not None:
            for el in doc.xpath('//*[@class]'):
                classes = filter(None, el.attrib['class'].split())
                if el.tag in self.allow_classes:
                    allowed = self.allow_classes[el.tag]
                    condition = allowed if callable(allowed) else \
                            (lambda cls: cls in allowed)
                    classes = filter(condition, classes)
                else:
                    classes = []

                if classes:
                    el.attrib['class'] = ' '.join(classes)
                else:
                    el.attrib.pop('class')


        for callback in self.dom_callbacks:
            callback(doc)

        if self.wrap_inline_tags is not False and self.tags_to_wrap:
            self.clean_top(doc)

        if self.split_paragraphs_by_br:
            self.remove_brs_from_pars(doc)

        for tag in self.drop_empty_tags:
            for el in doc.xpath('//'+tag):
                if not el.attrib and self.is_element_empty(el):
                    el.drop_tree()


def sanitize(value, **kwargs):
    doc = html.fragment_fromstring(value, create_parent=True)
    Cleaner(**kwargs)(doc)
    clean = html.tostring(doc, encoding='utf-8').decode('utf-8')
    return clean.split('>', 1)[1].rsplit('<', 1)[0]



from __future__ import absolute_import

import sys, functools, inspect


def return_locals(func):
    '''Modifies decorated function to return its locals'''

    @functools.wraps(func)
    def wrap(*args, **kwargs):
        frames = []

        def tracer(frame, event, arg): # pragma: no cover
            # coverage does not work in this function because the tracer
            # is deactivated here
            frames.append(frame)
            sys.settrace(old_tracer)
            if old_tracer is not None:
                return old_tracer(frame, event, arg)

        old_tracer = sys.gettrace()
        # tracer is activated on next call, return or exception
        sys.settrace(tracer)
        try:
            func(*args, **kwargs)
        finally:
            sys.settrace(old_tracer)
        assert len(frames) == 1
        argspec = inspect.getargspec(func)
        argnames = list(argspec.args)
        if argspec.varargs is not None:
            argnames.append(argspec.varargs)
        if argspec.keywords is not None:
            argnames.append(argspec.keywords)
        return {name: value for name, value in frames.pop(0).f_locals.items()
                if name not in argnames}

    return wrap



# -*- coding: utf-8 -*-

import logging

import six
import cgi
import re
from . import convs, widgets
from ..utils import cached_property
from collections import OrderedDict
from .perms import FieldPerm

logger = logging.getLogger(__name__)

__all__ = ['BaseField', 'Field', 'FieldBlock', 'FieldSet', 'FieldList', 'FileField']


class BaseField(object):
    '''
    Simple container class which ancestors represents various parts of Form.

    Encapsulates converter, various fields attributes, methods for data 
    access control
    '''

    # obsolete parameters from previous versions
    _obsolete = frozenset(['default', 'get_default', 'template', 'media',
                           'render_type', 'render', 'required'])

    #: :class:`FieldPerm` instance determining field's access permissions.
    #: Can be set by field inheritance or throught constructor.
    perm_getter = FieldPerm()

    # defaults
    #: :class:`Converter` instance determining field's convertation method
    conv = convs.Char()
    #: :class:`Widget` instance determining field's render method
    widget = widgets.TextInput
    #: Unicode label of the field
    label = None
    #: Short description of the field
    hint = None

    help = ''

    def __init__(self, name, conv=None, parent=None, permissions=None, **kwargs):
        if self._obsolete & set(kwargs):
            raise TypeError(
                    'Obsolete parameters are used: {}'.format(
                                list(self._obsolete & set(kwargs))))
        kwargs.update(
            parent=parent,
            name=name,
            conv=(conv or self.conv)(field=self),
            widget=(kwargs.get('widget') or self.widget)(field=self),
        )
        if permissions is not None:
            kwargs['perm_getter'] = FieldPerm(permissions)
        self._init_kwargs = kwargs
        self.__dict__.update(kwargs)

    def __call__(self, **kwargs):
        '''
        Creates current object's copy with extra constructor arguments passed.
        '''
        params = dict(self._init_kwargs, **kwargs)
        return self.__class__(**params)

    @property
    def multiple(self):
        return self.conv.multiple

    @property
    def env(self):
        return self.parent.env

    @property
    def form(self):
        return self.parent.form

    @property
    def input_name(self):
        '''
        Name of field's input element generated in account to possible
        nesting of fields. The input name is to be used in templates as value
        of Input (Select, etc) element's Name attribute and Label element's For
        attribute.
        '''
        return self.parent.prefix + self.name

    @property
    def error(self):
        '''
        String description of validation error in this field during last accept.

        `None` if there is no error.
        '''
        return self.form.errors.get(self.input_name)

    @property
    def help_message(self):
        return self.help or self.form.get_help(self.input_name)

    @cached_property
    def clean_value(self):
        '''
        Current field's converted value from form's python_data.
        '''
        # XXX cached_property is used only for set initial state
        #     this property should be set every time field data
        #     has been changed, for instance, in accept method
        python_data = self.parent.python_data
        if self.name in python_data:
            return python_data[self.name]
        return self.get_initial()

    @property
    def id(self):
        if self.form.id:
            # We use template names in list to replace, so we must use it here to
            # insure unique IDs.
            return '{}-{}'.format(self.form.id, self.input_name)
        return self.input_name

    def from_python(self, value):
        return self.conv.from_python(value)

    @cached_property
    def permissions(self):
        '''
        Field's access permissions. By default, is filled from perm_getter.
        '''
        return self.perm_getter.get_perms(self)

    @cached_property
    def writable(self):
        return 'w' in self.permissions

    @cached_property
    def readable(self):
        return 'r' in self.permissions

    @cached_property
    def field_names(self):
        return [self.name]

    def load_initial(self, initial, raw_data):
        value = initial.get(self.name, self.get_initial())
        self.set_raw_value(raw_data,
                           self.from_python(value))
        return {self.name: value}

    def __repr__(self):
        args = ', '.join([k+'='+repr(v)
                          for k, v in self._init_kwargs.items()
                          if k not in ['widget', 'conv', 'parent']])
        return '{}({})'.format(self.__class__.__name__, args)


class Field(BaseField):
    '''
    Atomic field
    '''

    conv = convs.Char()
    _null_value = ''

    def get_initial(self):
        if hasattr(self, 'initial'):
            return self.initial
        if self.multiple:
            return []
        return None

    @property
    def raw_value(self):
        if self.multiple:
            return self.form.raw_data.getall(self.input_name)
        else:
            return self.form.raw_data.get(self.input_name, '')

    def set_raw_value(self, raw_data, value):
        if self.multiple:
            try:
                del raw_data[self.input_name]
            except KeyError:
                pass
            for v in value:
                raw_data.add(self.input_name, v)
        else:
            raw_data[self.input_name] = value

    def _check_value_type(self, values):
        if not self.multiple:
            values = [values]
        for value in values:
            if not isinstance(value, six.string_types):
                self.form.errors[self.input_name] = 'Given value has incompatible type'
                return False
        return True

    def accept(self):
        '''Extracts raw value from form's raw data and passes it to converter'''
        value = self.raw_value
        if not self._check_value_type(value):
            # XXX should this be silent or TypeError?
            value = [] if self.multiple else self._null_value
        self.clean_value = self.conv.accept(value)
        return {self.name: self.clean_value}


class AggregateField(BaseField):

    @property
    def python_data(self):
        '''Representation of aggregate value as dictionary.'''
        try:
            value = self.clean_value
        except LookupError:
            # XXX is this necessary?
            value = self.get_initial()
        return self.from_python(value)


class FieldSet(AggregateField):
    '''
    Container field aggregating a couple of other different fields
    '''
    conv = convs.Converter()
    widget = widgets.FieldSetWidget()
    fields = []

    def __init__(self, name, conv=None, fields=None, **kwargs):
        fields = fields if fields is not None else self.fields
        if kwargs.get('parent'):
            conv = (conv or self.conv)(field=self)
            fields = [field(parent=self) for field in fields]
        kwargs.update(
            name=name,
            conv=conv,
            fields=fields,
        )
        BaseField.__init__(self, **kwargs)

    @property
    def prefix(self):
        return self.input_name+'.'

    def get_field(self, name):
        names = name.split('.', 1)
        for field in self.fields:
            if isinstance(field, FieldBlock):
                result = field.get_field(name)
                if result is not None:
                    return result
            if field.name == names[0]:
                if len(names) > 1:
                    return field.get_field(names[1])
                return field
        return None

    def get_initial(self):
        field_names = sum([x.field_names for x in self.fields], [])
        result = dict((name, self.get_field(name).get_initial())
                      for name in field_names)
        return self.conv.accept(result, silent=True)

    def set_raw_value(self, raw_data, value):
        # fills in raw_data multidict, resulting keys are field's absolute names
        assert isinstance(value, dict), \
                'To set raw value on {!r} need dict, got {!r}'\
                        .format(self.input_name, value)
        if not value:
            # Field set can be optional
            return
        field_names = sum([x.field_names for x in self.fields], [])
        for field_name in field_names:
            subvalue = value[field_name]
            field = self.get_field(field_name)
            field.set_raw_value(raw_data, field.from_python(subvalue))

    def accept(self):
        '''
        Accepts all children fields, collects resulting values into dict and
        passes that dict to converter.

        Returns result of converter as separate value in parent `python_data`
        '''
        result = dict(self.python_data)
        for field in self.fields:
            if field.writable:
                result.update(field.accept())
            else:
                # readonly field
                field.set_raw_value(self.form.raw_data,
                                    field.from_python(result[field.name]))
        self.clean_value = self.conv.accept(result)
        return {self.name: self.clean_value}


class FieldBlock(FieldSet):
    '''
    Anonymous FieldSet, values of one are accepted as they are children 
    of FieldBlock's parent.

    FieldBlock is used to logically organize fields and do validation
    of group of fields without naming that group and without dedicating 
    result of accept to separate object.
    '''

    conv = convs.FieldBlockConv()
    widget = widgets.FieldBlockWidget()
    prefix = ''

    def __init__(self, title, fields=[], **kwargs):
        kwargs.update(
            title=title,
            fields=fields,
        )
        kwargs.setdefault('name', '') # XXX generate unique name
        FieldSet.__init__(self, **kwargs)

    @cached_property
    def prefix(self):
        return self.parent.prefix

    def accept(self):
        '''
        Acts as `Field.accepts` but returns result of every child field 
        as value in parent `python_data`.
        '''
        result = FieldSet.accept(self)
        self.clean_value = result[self.name]
        return self.clean_value

    def load_initial(self, initial, raw_data):
        result = {}
        for field in self.fields:
            result.update(field.load_initial(initial, raw_data))
        return result

    @cached_property
    def field_names(self):
        result = []
        for field in self.fields:
            result += field.field_names
        return result

    @property
    def python_data(self):
        # we need only subfield values in python data
        result = {}
        for field_name in self.field_names:
            if field_name in self.parent.python_data:
                result[field_name] = self.parent.python_data[field_name]
        return result


class FieldList(AggregateField):
    '''
    Container aggregating an ordered set of similar fields
    '''

    order = True
    conv = convs.List()
    widget = widgets.FieldListWidget()
    _digit_re = re.compile('\d+$')

    def __init__(self, name, conv=None, field=Field(None),
                 parent=None, **kwargs):
        if parent:
            conv = (conv or self.conv)(field=self)
            field = field(parent=self)
        kwargs.update(
            parent=parent,
            name=name,
            conv=conv,
            field=field,
        )
        BaseField.__init__(self, **kwargs)

    @property
    def prefix(self):
        # NOTE: There was '-' instead of '.' and get_field('list-1') was broken
        return self.input_name+'.'

    def get_initial(self):
        return []

    def get_field(self, name):
        names = name.split('.', 1)
        if not self._digit_re.match(names[0]):
            # XXX is this needed?
            return None
        field = self.field(name=names[0])
        if len(names) > 1:
            return field.get_field(names[1])
        return field

    @property
    def indices_input_name(self):
        return self.input_name+'-indices'

    def accept(self):
        old = self.python_data
        result = OrderedDict()
        for index in self.form.raw_data.getall(self.indices_input_name):
            try:
                #XXX: we do not convert index to int, just check it.
                #     is it good idea?
                int(index)
            except ValueError:
                logger.warning('Got incorrect index from form: %r', index)
                continue
            #TODO: describe this
            field = self.field(name=str(index))
            if not field.writable:
                # readonly field
                if index in old:
                    result[field.name] = old[field.name]
            else:
                result.update(field.accept())
        self.clean_value = self.conv.accept(result)
        return {self.name: self.clean_value}

    def set_raw_value(self, raw_data, value):
        indices = []
        for index in range(1, len(value)+1):
            index = str(index)
            subvalue = value[index]
            subfield = self.field(name=index)
            subfield.set_raw_value(raw_data, subfield.from_python(subvalue))
            indices.append(index)
        try:
            del raw_data[self.indices_input_name]
        except KeyError:
            pass
        for index in indices:
            raw_data.add(self.indices_input_name, index)


class FileField(Field):
    '''
    The simpliest file field
    '''

    _null_value = None
    conv = convs.SimpleFile()

    def set_raw_value(self, raw_data, value):
        pass

    def _check_value_type(self, values):
        if not self.multiple:
            values = [values]
        for value in values:
            if not isinstance(value, cgi.FieldStorage) and \
               value and \
               not hasattr(value, 'read'): # XXX is this right?
                self.form.errors[self.input_name] = 'Given value is not file'
                return False
        return True




# -*- coding: utf-8 -*-

import os, sys, time, errno, six, io


def is_running(pid):
    '''Returns True if process with PID `pid` is running. Current user must
    have permission to access process information.'''
    try:
        os.kill(pid, 0)
    except OSError as exc:
        if exc.errno == errno.ESRCH:
            return False
        raise
    return True


def terminate(pid, sig, timeout):
    '''Terminates process with PID `pid` and returns True if process finished
    during `timeout`. Current user must have permission to access process
    information.'''
    os.kill(pid, sig)
    start = time.time()
    while True:
        try:
            # This is requireed if it's our child to avoid zombie. Also
            # is_running() returns True for zombie process.
            _, status = os.waitpid(pid, os.WNOHANG)
        except OSError as exc:
            if exc.errno != errno.ECHILD: # pragma: nocover
                raise
        else:
            if status:
                return True
        if not is_running(pid):
            return True
        if time.time()-start>=timeout:
            return False
        time.sleep(0.1)


def safe_makedirs(*files):
    for filename in files:
        dirname = os.path.dirname(filename)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)


def doublefork(pidfile, logfile, cwd, umask): # pragma: nocover
    '''Daemonize current process.
    After first fork we return to the shell and removing our self from
    controling terminal via `setsid`.
    After second fork we are not session leader any more and cant get
    controlling terminal when opening files.'''
    try:
        if os.fork():
            os._exit(os.EX_OK)
    except OSError as e:
        sys.exit('fork #1 failed: ({}) {}'.format(e.errno, e.strerror))
    os.setsid()
    os.chdir(cwd)
    os.umask(umask)
    try:
        if os.fork():
            os._exit(os.EX_OK)
    except OSError as e:
        sys.exit('fork #2 failed: ({}) {}'.format(e.errno, e.strerror))
    if logfile is not None:
        si = open('/dev/null')
        if six.PY2:
            so = open(logfile, 'a+', 0)
        else:
            so = io.open(logfile, 'ab+', 0)
            so = io.TextIOWrapper(so, write_through=True, encoding="utf-8")

        os.dup2(si.fileno(), 0)
        os.dup2(so.fileno(), 1)
        os.dup2(so.fileno(), 2)
        sys.stdin = si
        sys.stdout = sys.stderr = so
    with open(pidfile, 'w') as f:
        f.write(str(os.getpid()))



# -*- coding: utf-8 -*-
import six
import logging
from importlib import import_module
from sqlalchemy import create_engine


def multidb_binds(databases, package=None, engine_params=None):
    '''Creates dictionary to be passed as `binds` parameter to
    `sqlalchemy.orm.sessionmaker()` from dictionary mapping models module name
    to connection URI that should be used for these models. Models module must
    have `metadata` attribute. `package` when set must be a package or package
    name for all models modules.'''
    engine_params = engine_params or {}
    if not (package is None or isinstance(package, six.string_types)):
        package = getattr(package, '__package__', None) or package.__name__
    binds = {}
    for ref, uri in databases.items():
        md_ref = '.'.join(filter(None, [package, ref]))
        md_module = import_module(md_ref)
        try:
            metadata = md_module.metadata
        except AttributeError:
            raise ImportError(
                'Cannot import name metadata from module {}'.format(md_ref))
        engine = create_engine(uri, **engine_params)
        # Dot before [name] is required to allow setting logging level etc. for
        # all them at once.
        engine.logger = logging.getLogger('sqlalchemy.engine.[%s]' % ref)
        for table in metadata.sorted_tables:
            binds[table] = engine
    return binds



# -*- coding: utf-8 -*-
"""
    Originally from werkzeug.urls

    :copyright: (c) 2014 by the Werkzeug Team, see AUTHORS for more details.
    :license: BSD
"""
from six import text_type
from six.moves.urllib.parse import quote

_hexdigits = '0123456789ABCDEFabcdef'
_hextobyte = dict(
    ((a + b).encode(), int(a + b, 16))
    for a in _hexdigits for b in _hexdigits
)

def url_unquote(string, unsafe=''):
    """URL decode a single string with a given encoding.  If the charset
    is set to `None` no unicode decoding is performed and raw bytes
    are returned.

    :param string: the string to unquote.
    """
    rv = _unquote_to_bytes(string, unsafe)
    return repercent_broken_unicode(rv)


def _unquote_to_bytes(string, unsafe=''):
    if isinstance(string, text_type):
        string = string.encode('utf-8')
    if isinstance(unsafe, text_type):
        unsafe = unsafe.encode('utf-8')
    unsafe = frozenset(bytearray(unsafe))
    bits = iter(string.split(b'%'))
    result = bytearray(next(bits, b''))
    for item in bits:
        try:
            char = _hextobyte[item[:2]]
            if char in unsafe:
                raise KeyError()
            result.append(char)
            result.extend(item[2:])
        except KeyError:
            result.extend(b'%')
            result.extend(item)
    return bytes(result)


def repercent_broken_unicode(path):
    """
    As per section 3.2 of RFC 3987, step three of converting a URI into an IRI,
    we need to re-percent-encode any octet produced that is not part of a
    strictly legal UTF-8 octet sequence.
    """
    # originally from django.utils.encoding
    while True:
        try:
            return path.decode('utf-8')
        except UnicodeDecodeError as e:
            repercent = quote(path[e.start:e.end], safe=b"/#%[]=:;$&()+,!?*@'~")
            path = path[:e.start] + repercent.encode('ascii') + path[e.end:]


def uri_to_iri_parts(path, query, fragment):
    r"""
    Converts a URI parts to corresponding IRI parts in a given charset.

    Examples for URI versus IRI:

    :param path: The path of URI to convert.
    :param query: The query string of URI to convert.
    :param fragment: The fragment of URI to convert.
    """
    path = url_unquote(path, '%/;?')
    query = url_unquote(query, '%;/?:@&=+,$#')
    fragment = url_unquote(fragment, '%;/?:@&=+,$#')
    return path, query, fragment




# -*- coding: utf-8 -*-

import os
import hashlib
import logging
import binascii
from webob.exc import HTTPSeeOther

logger = logging.getLogger(__name__)


from iktomi import web
from iktomi.forms import *
from iktomi.utils.i18n import N_
from iktomi.storage import LocalMemStorage


def encrypt_password(raw_password, algorithm='sha1', salt=None):
    """
    Returns a string of the hexdigest of the given plaintext password and salt
    using the given algorithm ('md5', 'sha1' or other supported by hashlib).
    """
    if salt is None:
        salt = binascii.hexlify(os.urandom(3))[:5]
    else:
        salt = salt.encode('utf-8')

    raw_password = raw_password.encode('utf-8')
    hash = hashlib.new(algorithm, salt+raw_password).hexdigest()
    return '{}${}${}'.format(algorithm, salt.decode('utf-8'), hash)


def check_password(raw_password, enc_password):
    """
    Returns a boolean of whether the raw_password was correct. Handles
    encryption formats behind the scenes.
    """
    algo, salt, hsh = enc_password.split('$')
    return enc_password == encrypt_password(raw_password, algorithm=algo,
                                            salt=salt)


class LoginForm(Form):
    fields = (
        Field('login', convs.Char(),
              label=N_('Username')),
        Field('password', convs.Char(),
              widget=widgets.PasswordInput(),
              label=N_(u'Password')))


class CookieAuth(web.WebHandler):

    def __init__(self, get_user_identity, identify_user, storage=None,
                 cookie_name='auth', login_form=LoginForm,
                 crash_without_storage=True, expire_time=0):
        self.get_user_identity = get_user_identity
        self.identify_user = identify_user
        self._cookie_name = cookie_name
        self._login_form = login_form
        self.storage = LocalMemStorage() if storage is None else storage
        self.crash_without_storage = crash_without_storage
        self.expire_time = expire_time

    def cookie_auth(self, env, data):
        user = None
        if self._cookie_name in env.request.cookies:
            key = env.request.cookies[self._cookie_name]
            storage_key = self._cookie_name + ':' + key
            user_identity = self.storage.get(storage_key)
            if user_identity is not None:
                user = self.identify_user(env, user_identity)
                self.storage.set(storage_key, user_identity, self.expire_time)
        logger.debug('Authenticated: %r', user)
        env.user = user
        try:
            result = self.next_handler(env, data)
        finally:
            del env.user
        return result
    __call__ = cookie_auth

    def login_identity(self, user_identity, response=None, path='/'):
        key = binascii.hexlify(os.urandom(10)).decode('ascii')
        response = web.Response() if response is None else response
        response.set_cookie(self._cookie_name, key, path=path)
        storage_key = self._cookie_name+':'+key
        if not self.storage.set(storage_key, str(user_identity),
                                self.expire_time):
            logger.warning('storage "%r" is unreachable', self.storage)
            if self.crash_without_storage:
                raise Exception(
                        'Storage {!r} is gone or down'.format(self.storage))
        return response

    def logout_user(self, request, response):
        if self._cookie_name in request.cookies:
            response.delete_cookie(self._cookie_name)
            key = request.cookies[self._cookie_name]
            if key is not None:
                if not self.storage.delete(self._cookie_name + ':' + key):
                    logger.warning('storage "%r" is unreachable', self.storage)

    def login(self, template='login'):
        '''
        This property will return component which will handle login requests.

            auth.login(template='login.html')
        '''
        def _login(env, data):
            form = self._login_form(env)
            next = env.request.GET.get('next', '/')
            login_failed = False
            if env.request.method == 'POST':
                if form.accept(env.request.POST):
                    user_identity = self.get_user_identity(
                                                env, **form.python_data)
                    if user_identity is not None:
                        response = HTTPSeeOther(location=next)
                        return self.login_identity(user_identity, response)
                    login_failed = True
            data.form = form
            data.login_failed = login_failed
            data.login_url = env.root.login.as_url.qs_set(next=next)
            return env.template.render_to_response(template, data.as_dict())
        return web.match('/login', 'login') | _login

    def logout(self, redirect_to='/'):
        '''
        This property will return component which will handle logout requests.
        It only handles POST requests and do not display any rendered content.
        This handler deletes session id from `storage`. If there is no
        session id provided or id is incorrect handler silently redirects to
        login url and does not throw any exception.
        '''
        def _logout(env, data):
            location = redirect_to
            if location is None and env.request.referer:
                location = env.request.referer
            elif location is None:
                location = '/'
            response = HTTPSeeOther(location=str(location))
            self.logout_user(env.request, response)
            return response
        return web.match('/logout', 'logout') | web.method('post') | _logout


@web.request_filter
def auth_required(env, data, next_handler):
    if getattr(env, 'user', None) is not None:
        return next_handler(env, data)
    response = web.Response(status=303)
    response.headers['Location'] = str(
                env.root.login.as_url.qs_set(next=env.request.path_info))
    return response


class SqlaModelAuth(CookieAuth):

    def __init__(self, model, storage=None, login_field='login',
                 password_field='password', **kwargs):
        self._model = model
        self._login_field = login_field
        self._password_field = password_field
        CookieAuth.__init__(self, self.get_user_identity, self.identify_user,
                            storage=storage, **kwargs)

    def get_query(self, env, login):
        model = self._model
        login_field = getattr(model, self._login_field)
        return env.db.query(model).filter(login_field==login)

    def get_user_identity(self, env, login, password):
        user = self.get_query(env, login).first()
        if user is not None:
            stored_password = getattr(user, self._password_field)
            if check_password(password, stored_password):
                return user.id
        return None

    def identify_user(self, env, user_identity):
        return env.db.query(self._model).get(user_identity)



# -*- coding: utf-8 -*-
import six
if six.PY2:
    from urllib import quote, unquote
else:# pragma: no cover
    from urllib.parse import quote, unquote

import re
import logging
from .url_converters import default_converters, ConvertError

logger = logging.getLogger(__name__)


def urlquote(value):
    if isinstance(value, six.integer_types):
        value = six.text_type(value)
    return quote(value.encode('utf-8'))


class UrlBuildingError(Exception): pass


_split_pattern = re.compile(r'(<[^<]*>)')

#NOTE: taken from werkzeug
_converter_pattern = re.compile(r'''^<
        (?:
            (?P<converter>[a-zA-Z_][a-zA-Z0-9_]+)   # converter name
            (?:\((?P<args>.*?)\))?                  # converter args
            \:                                      # delimiter
        )?
        (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
        >$''', re.VERBOSE | re.U)

_static_url_pattern = re.compile(r'^[^<]*?$')

def construct_re(url_template, match_whole_str=False, converters=None,
                 default_converter='string', anonymous=False):
    '''
    url_template - str or unicode representing template

    Constructed pattern expects urlencoded string!

    returns  (compiled re pattern, 
              dict {url param name: [converter name, converter args (str)]},
              list of (variable name, converter name, converter args name))

    If anonymous=True is set, regexp will be compiled without names of variables.
    This is handy for example, if you want to dump an url map to JSON.
    '''
    # needed for reverse url building (or not needed?)
    builder_params = []
    # found url params and their converters
    url_params = {}
    result = r'^'
    parts = _split_pattern.split(url_template)
    for i, part in enumerate(parts):
        is_url_pattern = _static_url_pattern.match(part)
        if is_url_pattern:
            #NOTE: right order:
            #      - make part str if it was unicode
            #      - urlquote part
            #      - escape all specific for re chars in part
            result += re.escape(urlquote(part))
            builder_params.append(part)
            continue
        is_converter = _converter_pattern.match(part)
        if is_converter:
            groups = is_converter.groupdict()
            converter_name = groups['converter'] or default_converter
            conv_object = init_converter(converters[converter_name],
                                         groups['args'])
            variable = groups['variable']
            builder_params.append((variable, conv_object))
            url_params[variable] = conv_object
            if anonymous:
                result += conv_object.regex
            else:
                result += '(?P<{}>{})'.format(variable, conv_object.regex)
            continue
        raise ValueError('Incorrect url template {!r}'.format(url_template))
    if match_whole_str:
        result += '$'
    return re.compile(result), url_params, builder_params


def init_converter(conv_class, args):
    if args:
        #XXX: taken from werkzeug
        storage = type('_Storage', (), {'__getitem__': lambda s, x: x})()
        args, kwargs = eval(u'(lambda *a, **kw: (a, kw))({})'.format(args),
                            {}, storage)
        return conv_class(*args, **kwargs)
    return conv_class()


class UrlTemplate(object):

    def __init__(self, template, match_whole_str=True, converters=None,
                 default_converter='string'):
        self.template = template
        self.match_whole_str = match_whole_str
        self._allowed_converters = self._init_converters(converters)
        self._pattern, self._url_params, self._builder_params = \
            construct_re(template,
                         match_whole_str=match_whole_str,
                         converters=self._allowed_converters,
                         default_converter=default_converter)

    def match(self, path, **kw):
        '''
        path - str (urlencoded)
        '''
        m = self._pattern.match(path)
        if m:
            kwargs = m.groupdict()
            # convert params
            for url_arg_name, value_urlencoded in kwargs.items():
                conv_obj = self._url_params[url_arg_name]
                unicode_value = unquote(value_urlencoded)
                if isinstance(unicode_value, six.binary_type):
                    # XXX ??
                    unicode_value = unicode_value.decode('utf-8', 'replace')
                try:
                    kwargs[url_arg_name] = conv_obj.to_python(unicode_value, **kw)
                except ConvertError as err:
                    logger.debug('ConvertError in parameter "%s" '
                                 'by %r, value "%s"',
                                 url_arg_name,
                                 err.converter.__class__,
                                 err.value)
                    return None, {}
            return m.group(), kwargs
        return None, {}

    def __call__(self, **kwargs):
        'Url building with url params values taken from kwargs. (reverse)'
        result = ''
        for part in self._builder_params:
            if isinstance(part, tuple):
                var, conv_obj = part
                try:
                    value = kwargs[var]
                except KeyError:
                    if conv_obj.default is not conv_obj.NotSet:
                        value = conv_obj.default
                    else:
                        raise UrlBuildingError('Missing argument for '
                                               'URL builder: {}'.format(var))
                result += conv_obj.to_url(value)
            else:
                result += part
        # result - unicode not quotted string
        return result

    def _init_converters(self, converters):
        convs = default_converters.copy()
        if converters is not None:
            convs.update(converters)
        return convs

    def __eq__(self, other):
        return self.template == other.template and \
               self.match_whole_str == other.match_whole_str

    def __repr__(self):
        return '{}({!r}, match_whole_str={!r})'.format(
                self.__class__.__name__, self.template,
                self.match_whole_str)



# -*- coding: utf-8 -*-

__all__ = ['WebHandler', 'cases', 'request_filter']

import logging
import functools

from copy import copy
from webob import Response

logger = logging.getLogger(__name__)


def is_chainable(handler):
    while isinstance(handler, WebHandler):
        if not hasattr(handler, '_next_handler'):
            return True
        handler = handler._next_handler
    return False

def respond(response):
    def response_wrapper(env, data):
        return response
    return response_wrapper

def prepare_handler(handler):
    if isinstance(handler, Response):
        return respond(handler)
    elif isinstance(handler, type) and \
         issubclass(handler, Response):
        return respond(handler())
    return handler


class WebHandler(object):
    '''Base class for all request handlers.'''

    def __or__(self, next_handler):
        '''
        Supports chaining handler after itself::

            WebHandlerSubclass() | another_handler
        '''
        # XXX in some cases copy count can be big
        #     for example, chaining something after a huge cases(..) handler
        #     causes a copy of each single nested handler.
        #     Sure, is bad idea to chain anything after big cases(..) anyway.
        h = self.copy()

        next_handler = prepare_handler(next_handler)
        if hasattr(self, '_next_handler'):
            h._next_handler = h._next_handler | next_handler
        else:
            h._next_handler = next_handler
        return h

    def _locations(self):
        next_handler = self.next_handler
        if isinstance(next_handler, WebHandler):
            return next_handler._locations()
        # we are last in chain
        return {}

    def __repr__(self):
        return '{}()'.format(self.__class__.__name__)

    def __call__(self, env, data):
        '''
        Subclasses should define __call__ with handler code. It is good style to
        give a name similar to handler's name to method and then make an alias
        to __call__::

            class MyHandler(WebHandler):

                def my_handler(self, env, data):
                    do_something(env, data)
                    return self.next_handler(env, data)
                __call__ = my_handler  # This method should be overridden 
                                       # in subclasses.
        '''
        raise NotImplementedError(
                '__call__ is not implemented in {!r}'.format(self))

    @property
    def next_handler(self):
        '''A handler, chained next to self'''
        if hasattr(self, '_next_handler'):
            return self._next_handler
        return lambda e, d: None

    def copy(self):
        '''
        Returns copy for the handler to make handlers reusable.
        Handlers are being copied automatically on chaining,
        so you do not need to do it manually.'''
        return copy(self)


class cases(WebHandler):
    # XXX bad docstring
    '''
    Handler incapsulating multiple routing branches and choosing one of them
    that matches current request::

        web.cases(
            web.match('/', 'index') | index,
            web.match('/contacts', 'contacts') | contacts,
            web.match('/about', 'about') | about,
        )'''

    def __init__(self, *handlers):
        self.handlers = [prepare_handler(x) for x in handlers]

    def __or__(self, next_handler):
        #cases needs to set next handler for each handler it keeps
        h = self.copy()
        h.handlers = [(handler | next_handler
                            if is_chainable(handler) 
                            else handler)
                      for handler in self.handlers]
        return h

    def cases(self, env, data):
        '''Calls each nested handler until one of them returns nonzero result.

        If any handler returns `None`, it is interpreted as 
        "request does not match, the handler has nothing to do with it and 
        `web.cases` should try to call the next handler".'''
        for handler in self.handlers:
            env._push()
            data._push()
            try:
                result = handler(env, data)
            finally:
                env._pop()
                data._pop()
            if result is not None:
                return result
    # for readable tracebacks
    __call__ = cases

    def _locations(self):
        locations = {}
        for handler in self.handlers:
            if isinstance(handler, WebHandler):
                handler_locations = handler._locations()
                for k, v in handler_locations.items():
                    if k in locations:
                        raise ValueError(
                                'Location "{}" already exists'.format(k))
                    locations[k] = v
        return locations

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__,
                               ', '.join(repr(h) for h in self.handlers))


class _FunctionWrapper3(WebHandler):
    '''
    Wrapper for handler represented by function 
    (3 args, old-style)
    '''

    def __init__(self, func):
        self.handler = func

    def function_wrapper(self, env, data):
        return self.handler(env, data, self.next_handler)
    __call__ = function_wrapper

    def __repr__(self):
        return '{}({!r})'.format(self.__class__.__name__, self.handler)


def request_filter(func):
    '''Decorator transforming function to regular WebHandler.
    This allows to chain other handlers after given.
    The next handler is passed as third argument into the wrapped function::

        @web.request_filter
        def wrapper(env, data, next_handler):
            do_something()
            result = next_handler(env, data)
            return do_something_else(result)

        wrapped_app = wrapper | handler
        '''
    return functools.wraps(func)(_FunctionWrapper3(func))




# -*- coding: utf-8 -*-
import six

from inspect import isclass
from datetime import datetime

__all__ = ['ConvertError', 'default_converters', 'Converter', 'String',
           'Integer', 'Any', 'Date']



class ConvertError(Exception):
    '''
    Converter should raise ConvertError if the given value does not match
    '''

    @property
    def converter(self):
        return self.args[0]

    @property
    def value(self):
        return self.args[1]


class Converter(object):
    '''A base class for urlconverters'''

    regex = '[.a-zA-Z0-9:@&+$,_%%-]+'
    class NotSet(object): pass
    default = NotSet

    def __init__(self, default=NotSet, regex=None):
        if not default is self.NotSet:
            self.default = default
        if regex is not None:
            self.regex = regex

    def to_python(self, value, env=None):
        '''
        Accepts unicode url part and returns python object.
        Should be implemented in subclasses
        '''
        raise NotImplementedError() # pragma: no cover

    def to_url(self, value):
        '''
        Accepts python object and returns unicode prepared to be used
        in url building.
        Should be implemented in subclasses
        '''
        raise NotImplementedError() # pragma: no cover


class String(Converter):
    '''
    Unquotes urlencoded string::
    '''

    min = 1
    max = None

    def __init__(self, min=None, max=None, **kwargs):
        Converter.__init__(self, **kwargs)
        self.min = min if min is not None else self.min
        self.max = max or self.max

    def to_python(self, value, env=None):
        self.check_len(value)
        return value

    def to_url(self, value):
        if six.PY3 and isinstance(value, bytes):
            raise TypeError() # pragma: no cover, safety check
        return six.text_type(value)

    def check_len(self, value):
        length = len(value)
        if length < self.min or self.max and length > self.max:
            raise ConvertError(self, value)


class Integer(Converter):
    '''
    Extracts integer value from url part.
    '''

    regex = '(?:[1-9]\d*|0)'

    def to_python(self, value, env=None):
        try:
            value = int(value)
        except ValueError:
            raise ConvertError(self, value)
        else:
            return value

    def to_url(self, value):
        if isinstance(value, six.string_types):
            # sometimes it is useful to build fake urls with placeholders,
            # to be replaced in JS to real values
            # For example:
            #     root.item(id="REPLACEME")
            return value
        return str(int(value))


class Any(Converter):
    '''
    Checks if string value is in a list of allowed values and returns that value::

        web.match('/<any(yes,no,"probably, no",maybe):answer>')
    '''

    def __init__(self, *values, **kwargs):
        Converter.__init__(self, **kwargs)
        self.values = values

    def to_python(self, value, env=None):
        if value in self.values:
            return value
        raise ConvertError(self, value)

    def to_url(self, value):
        if six.PY3 and isinstance(value, bytes):
            raise TypeError() # pragma: no cover, safety check
        return six.text_type(value)


class Date(Converter):
    '''
    Converts string to datetime by strptime using given format::

        web.match('/<date(format="%Y.%m.%d"):date>')
    '''

    format = "%Y-%m-%d"

    def __init__(self, format=None, **kwargs):
        Converter.__init__(self, **kwargs)
        if format is not None:
            self.format = format

    def to_python(self, value, env=None):
        try:
            return datetime.strptime(value, self.format).date()
        except ValueError:
            raise ConvertError(self, value)

    def to_url(self, value):
        if isinstance(value, six.string_types):
            # sometimes it is useful to build fake urls with placeholders,
            # to be replaced in JS to real values
            # For example:
            #     root.item(id="REPLACEME")
            return value
        return value.strftime(self.format)


default_converters = {'string': String,
                      'int': Integer,
                      'any': Any,
                      'date': Date}

# assert all defined converters are registered
for item in list(globals().values()):
    if isclass(item) and \
       issubclass(item, Converter) and \
       not item is Converter:
        assert item in default_converters.values(), item




import six
from sqlalchemy.types import TypeDecorator, String, Text


class StringList(TypeDecorator):

    impl = String

    def process_bind_param(self, value, dialect):
        if value is not None:
            return ','.join(value)

    def process_result_value(self, value, dialect):
        if value is not None:
            return [x for x in value.split(',') if x]


class IntegerList(TypeDecorator):

    impl = String

    def process_bind_param(self, value, dialect):
        if value is not None:
            return ','.join(str(item) for item in value)

    def process_result_value(self, value, dialect):
        if value is not None:
            return [int(item) for item in value.split(',') if item]


try:
    from jinja2 import Markup
except ImportError: # pragma: no cover
    Markup = None


class HtmlBase(TypeDecorator):
    '''Base class for HTML markup types (safe to render in template)'''

    markup_class = Markup

    def process_result_value(self, value, dialect):
        if value is not None:
            return self.markup_class(value)

    def process_bind_param(self, value, dialect):
        if value is not None:
            return six.text_type(value)


class Html(HtmlBase):
    '''
    Factory class for HTML type. Usage:
        Column(Html(Text))
        Column(Html(String(1000)))
        Column(Html(BigText, markup_class=SomeWrapperClass))
    '''

    def __init__(self, _impl, markup_class=None):
        # Callable is useful to be able to pass type classes as well as
        # type objects, for example Html(String) and Html(String(100))
        if callable(_impl):
            _impl = _impl()
        self.impl = _impl
        self.markup_class = markup_class or self.markup_class
        # Don't call base class' __init__ since we reimplemented it in a
        # different way.


# Compatibility classes. Deprecate them?

class HtmlString(HtmlBase):
    impl = String

class HtmlText(HtmlBase):
    impl = Text



from iktomi.db.sqla.files import * # pragma: no cover
import warnings # pragma: no cover
warnings.warn('iktomi.unstable.db.sqla.files is deprecated. Use ' # pragma: no cover
              'iktomi.db.sqla.files instead',
              category=DeprecationWarning)



# -*- coding: utf-8 -*-
'''
This module contains basic form access control classes.

You can set permissions to form and fields like this::

    class SampleForm(form.Form):
       permissions = 'rwcx'
       fields=[fields.Field('input',
                            permissions='rw')]

or define your own permission logic::

    class SampleForm(form.Form):
       permissions = 'rwcx'
       fields=[fields.Field('input',
                            perm_getter=UserBasedPerm({'admin': 'rw',
                                                       'user': 'r'}))]


or pass form permissions to constructor::

    >>> form = SampleForm(env, data, permissions='rw')

To access current permissions set you can use field's :attr:`permissions`
property:

    >>> form.get_field('input').permissions
    set(['r', 'w'])

'''

DEFAULT_PERMISSIONS = set('rwc')


class FieldPerm(object):
    '''
    Default permission getter for Field objects

    Ancestor should override the :meth:`check` method. They can use field.env
    to get any values from outside. For example::

        class RoleBased(FieldPerm):
            def __init__(self, role_perms):
                self.role_perms = role_perms

            def check(self, field):
                user = field.env.user
                perms = set(self.role_perms.get('*', ''))
                for role in user.roles:
                    perms.update(self.role_perms.get(role, ''))
                return perms
    '''

    permissions = None

    def __init__(self, permissions=None):
        if permissions is not None:
            self.permissions = set(permissions)

    def get_perms(self, obj):
        '''
        Returns combined Environment's and object's permissions.
        Resulting condition is intersection of them.
        '''
        return self.available(obj) & self.check(obj)

    def available(self, field):
        '''
        Returns permissions according environment's limiting condition.
        Determined by object's context

        Allows only field's parents' permissions
        '''
        return field.parent.permissions

    def check(self, field):
        '''
        Returns permissions determined by object itself
        '''
        if self.permissions is None:
            return field.parent.permissions
        return self.permissions

    def __repr__(self):
        return '{}({})'.format(self.__class__.__name__, str(self.permissions))




from PIL import Image

# pillow 10 support
if hasattr(Image, "ANTIALIAS"):
    ANTIALIAS = Image.ANTIALIAS
else:
    ANTIALIAS = Image.Resampling.LANCZOS


class Resizer(object):

    def __init__(self, expand=False, filter=ANTIALIAS):
        self.expand = expand
        self.filter = filter

    def do__crop(self, img, transformation, params):
        return img.crop(params)

    def do__resize(self, img, transformation, params):
        return img.resize(params, self.filter)

    def new_size__resize(self, size, target_size, params):
        return params

    def new_size__crop(self, size, target_size, params):
        return (params[2] - params[0], params[3] - params[1])

    def transformations(self, size, target_size): # pragma: no cover
        '''
        Method describing transformations applied to the image,
        must be redefined in subclasses.

        Should return a list of JSON-serializable commands
        that are processed with transform method.
        Transformations can be dumped and applied on different place,
        for example, in Javascript
        '''
        raise NotImplementedError

    def transform(self, img, transformation, params):
        '''
        Apply transformations to the image.

        New transformations can be defined as methods::

            def do__transformationname(self, img, transformation, params):
                'returns new image with transformation applied'
                ...

            def new_size__transformationname(self, size, target_size, params):
                'dry run, returns a size of image if transformation is applied'
                ...
        '''
        # Transformations MUST be idempotent.
        # The limitation is caused by implementation of
        # image upload in iktomi.cms.
        # The transformation can be applied twice:
        # on image upload after crop (when TransientFile is created)
        # and on object save (when PersistentFile is created).
        method = getattr(self, 'do__' + transformation)
        return method(img, transformation, params)

    def get_target_size(self, size, target_size):
        transforms = self.transformations(size, target_size)
        for transformation, params in transforms:
            method = getattr(self, 'new_size__' + transformation)
            size = method(size, target_size, params)
        return size

    def __call__(self, img, target_size):
        transforms = self.transformations(img.size, target_size)
        for transformation, params in transforms:
            img = self.transform(img, transformation, params)
        return img


class ResizeFit(Resizer):

    '''
    Resizes an image in a way it fits to a given rectangle.

    :param expand: force expand image to the rectangle.
    '''

    def transformations(self, size, target_size):
        sw, sh = size
        tw, th = target_size
        if not self.expand and sw<=tw and sh<=th:
            return []

        if sw*th>sh*tw:
            h = sh*tw//sw
            return [('resize', (tw, h))]
        else:
            w = sw*th//sh
            return [('resize', (w, th))]


class ResizeCrop(Resizer):
    '''
    Crops image to the proportion of a given rectange and resizes it
    to the size of the rectangle.

    :param expand: force expand image to the rectangle.

    :param force: force image's proportion to be equal to the rectangle's proportion
    even if source image is smaller in any dimension than target size.
    '''

    def __init__(self, *args, **kwargs):
        self.force = kwargs.pop('force', False)
        Resizer.__init__(self, *args, **kwargs)
        assert not (self.force and self.expand)

    def transformations(self, size, target_size):
        sw, sh = size
        tw, th = target_size
        if not self.expand and not self.force and sw<=tw and sh<=th:
            return []

        if self.force and (sw<=tw or sh<=th):
            if sw*th>sh*tw:
                # crop right and left side
                tw, th = sh*tw//th, sh
            else:
                # crop upper and bottom side
                tw, th = sw, sw*th//tw

        transforms = []
        if sw*th>sh*tw:
            # crop right and left side
            if sh!=th and (sh>th or self.expand):
                w = sw*th//sh
                transforms.append(('resize', (w, th)))
                sw, sh = w, th
            if sw>tw:
                wd = (sw-tw)//2
                transforms.append(('crop', (wd, 0, tw+wd, sh)))
        else:
            # crop upper and bottom side
            if sw!=tw and (sw>tw or self.expand):
                h = sh*tw//sw
                transforms.append(('resize', (tw, h)))
                sw, sh = tw, h
            if sh>th:
                hd = (sh-th)//2
                transforms.append(('crop', (0, hd, sw, th+hd)))
        return transforms


class ResizeMixed(Resizer):

    '''
    Applies one of given rectangles depending on whether the image
    is vertical or horizontal.

    :param rate: multiplier to height to tune a proportion dividing
    horisontal images from verticals. Dy default is equal to 1 (1:1).
    '''

    def __init__(self, hor_resize, vert_resize, rate=1):
        self.hor_resize = hor_resize
        self.vert_resize = vert_resize
        self.rate = rate

    def get_resizer(self, size, target_size):
        '''Choose a resizer depending an image size'''
        sw, sh = size
        if sw >= sh * self.rate:
            return self.hor_resize
        else:
            return self.vert_resize

    def transformations(self, size, target_size):
        return self.get_resizer(size, target_size)\
                   .transformations(size, target_size)

    #def transform(self, img, *args):
    # XXX is this method needed?
    #    return self.get_resizer(img).transform(img, *args)

    def __call__(self, img, target_size):
        return self.get_resizer(img.size, target_size)(img, target_size)


class ResizeFixedWidth(Resizer):

    '''
    Resizes an image to a given width.

    :param expand: force expand image to target size.
    '''

    def transformations(self, size, target_size):
        sw, sh = size
        tw, th = target_size
        if not self.expand and sw<=tw:
            return []
        h = sh*tw//sw
        return [('resize', (tw, h))]


class ResizeFixedHeight(Resizer):

    '''
    Resizes an image to a given height.

    :param expand: force expand image to target size.
    '''

    def transformations(self, size, target_size):
        sw, sh = size
        tw, th = target_size
        if not self.expand and  sh<=th:
            return []
        w = sw*th//sh
        return [('resize', (w, th))]




import six
from iktomi.utils import cached_property
from iktomi.unstable.utils.functools import return_locals


class ModelFactories(object):

    def __init__(self):
        self.models = []
        self.i18n_models = []

    def get_constructor(self, func):
        return return_locals(func)

    def register(self, *base_names, **kwargs):
        lang = kwargs.pop('lang', False)
        langs = kwargs.pop('langs', None)
        assert not kwargs
        def decor(func):
            name = func.__name__
            constructor = self.get_constructor(func)
            if lang:
                self.i18n_models.append((name, constructor, base_names, langs))
            else:
                self.models.append((name, constructor, base_names))
            return constructor
        return decor

    def create_model(self, module, name, constructor, base_names):
        values = constructor(ModelsProxy(module, self))
        values['models'] = module
        return self.make_class(module, name, base_names, values)

    def make_class(self, module, name, base_names, values):
        is_string = lambda x: isinstance(x, six.string_types)

        bases = tuple(getattr(module, x) if is_string(x) else x
                      for x in base_names)
        cls = type(name, bases, values)
        cls.__module__ = module.__name__
        return cls

    def create_all(self, module, all_lang_modules=(), model_names=None):
        for name, constructor, base_names in self.models:
            if hasattr(module, name):
                continue
            if model_names is not None and name not in model_names:
                continue
            cls = self.create_model(module, name, constructor, base_names)
            setattr(module, name, cls)

        all_langs = [x.lang for x in all_lang_modules]
        for name, constructor, base_names, langs in self.i18n_models:
            if model_names is not None and name not in model_names:
                continue
            if langs is None:
                lang_modules = all_lang_modules
                langs = list(all_langs)
            else:
                lang_modules = [getattr(module, l) for l in langs]
            for lang_module in lang_modules:
                lang_name = lang_module._get_model_name(name)
                if hasattr(module, lang_name):
                    continue
                cls = self.create_model(lang_module, lang_name,
                                        constructor, base_names)
                # XXX proper name for this attribute?
                #     or may be, store it in external storage?
                cls._iktomi_langs = langs
                setattr(module, lang_name, cls)


class LangModelProxy(object):

    def __init__(self, module, langs, lang):
        self.module = module
        self.langs = langs
        self.lang = lang.lower()
        self.lang_upper = lang.title()

    @cached_property
    def main_lang(self):
        return self.langs[0]

    def _get_model_name(self, name):
        return name + self.lang_upper

    def __getattr__(self, name):
        lang_name = self._get_model_name(name)
        if hasattr(self.module, lang_name):
            return getattr(self.module, lang_name)
        return getattr(self.module, name)


class ModelsProxy(object):

    def __init__(self, models, factories):
        self.models = models
        self.i18n_model_names = [m[0] for m in factories.i18n_models]
        self.model_names = [m[0] for m in factories.models]

    def __getattr__(self, name):
        try:
            return getattr(self.models, name)
        except AttributeError:
            if hasattr(self.models, 'lang'):
                if name in self.i18n_model_names:
                    lang_name = self.models._get_model_name(name)
                    return PseudoModel(lang_name)
            if name in self.model_names:
                return PseudoModel(name)
        raise AttributeError("{} has not attribute {}".format(repr(self), name))


class PseudoModel(str):

    def __new__(cls, name, parent=None):
        qname = '.'.join([parent, name]) if parent else name
        return str.__new__(cls, qname)

    def __getattr__(self, name):
        return PseudoModel(name, self)




# -*- coding: utf-8 -*-

import os
import logging
logger = logging.getLogger(__name__)
from glob import glob
from ..web import Response, request_filter
from ..utils import cached_property

__all__ = ('Template',)


class TemplateError(Exception):
    '''Error raised in the case template can not be found'''
    pass


class Template(object):
    '''Proxy class managing a set of template engines'''

    def __init__(self, *dirs, **kwargs):
        self.globs = kwargs.get('globs', {})
        self.cache = kwargs.get('cache', False)
        self.dirs = []
        for d in dirs:
            self.dirs.append(d)
        self.engines = {}
        for template_type, engine in kwargs.get('engines', {}).items():
            self.engines[template_type] = engine

    def render(self, template_name, **kw):
        '''
        Given a template name and template vars.
        Searches a template file based on engine set, and renders it 
        with corresponding engine.
        Returns a string.
        '''
        logger.debug('Rendering template "%s"', template_name)
        vars = self.globs.copy()
        vars.update(kw)
        resolved_name, engine = self.resolve(template_name)
        return engine.render(resolved_name, **vars)

    def resolve(self, template_name):
        pattern = template_name
        if not os.path.splitext(template_name)[1]:
            pattern += '.*'
        for d in self.dirs:
            path = os.path.join(d, pattern)
            for file_name in glob(path):
                name, ext = os.path.splitext(file_name)
                template_type = ext[1:]
                if template_type in self.engines:
                    return file_name[len(d)+1:], self.engines[template_type]
        raise TemplateError(
                'Template or engine for template "{}" not found in '\
                'directories {!r}'.format(pattern, self.dirs))


class BoundTemplate(object):
    '''
    Object used to bound a `Template` object to iktomi environment.

    Usage::

        template = Template(...) # global var

        def environment(env, data, nxt):
            ...
            env.template = BoundTemplate(env, template)
    '''

    def __init__(self, env, template):
        self.template = template
        self.env = env

    def get_template_vars(self):
        '''
        Redefine this method to add extra template variables
        always available in loaded templates.

        Called on each template render'''
        return {}

    @cached_property
    def engines(self):
        return self.template.engines

    def _vars(self, __data, **kw):
        if hasattr(__data, 'as_dict'):
            d = __data.as_dict()
        elif __data is not None:
            d = dict(__data)
        else:
            d = {}
        d.update(kw)
        d.update(self.get_template_vars())
        return d

    def render(self, template_name, __data=None, **kw):
        '''Given a template name and template data.
        Renders a template and returns as string'''
        return self.template.render(template_name,
                                    **self._vars(__data, **kw))

    def render_to_response(self, template_name, __data,
                           content_type="text/html"):
        '''Given a template name and template data.
        Renders a template and returns `webob.Response` object'''
        resp = self.render(template_name, __data)
        return Response(resp,
                        content_type=content_type)



import sqlalchemy
from sqlalchemy.orm.query import Query
from sqlalchemy.orm.util import AliasedClass
from sqlalchemy.sql import ClauseElement, Join
from sqlalchemy.sql.selectable import FromGrouping
from sqlalchemy import cast, Boolean
from sqlalchemy.orm.util import _class_to_mapper
from sqlalchemy import event
from sqlalchemy import inspect
from packaging.version import parse as version


def init_filter_event(prop_name):
    @event.listens_for(Query, "before_compile", retval=True)
    def before_compile(query):
        """A query compilation rule that will add limiting criteria for every
        subclass of HasPrivate"""
        if query._execution_options.get("include_private", False):
            return query

        for ent in query.column_descriptions:
            entity = ent['entity']
            if entity is None:
                continue
            insp = inspect(ent['entity'])
            mapper = getattr(insp, 'mapper', None)
            if mapper and prop_name in dir(mapper.class_):
                query = query.enable_assertions(False).filter(
                    ent['entity'].public == True)

        return query
    
    return before_compile


class PublicQuery(Query):

    '''
    Filters all queries by publicity condition for each participating mapped
    class. Attribute "public" of mapped class (if present) should be either
    boolean column or @hybrid_property providing publicity criterion clause for
    the class and boolean (convertable to boolean) value for instance of the
    class.

    A version from recipe combined with our own vision
    http://www.sqlalchemy.org/trac/wiki/UsageRecipes/PreFilteredQuery

    UPD in v0.6
    Old recipe is outdated, we should use new one
    https://github.com/sqlalchemy/sqlalchemy/wiki/FilteredQuery

    Old functionality is untouched for compatibility
    '''

    property_name = 'public'

    # Sqlalchemy version should be lower than this version to be able
    # to handle all cases correctly
    # For newer versions we will use more modern and simpler solution
    # that cannot filter eager loading via joinedjoad
    max_sqla_version = '1.2'

    def __init__(self, *args, **kwargs):
        if version(sqlalchemy.__version__) >= version(self.max_sqla_version):
            init_filter_event(self.property_name)
        super(PublicQuery, self).__init__(*args, **kwargs)
        


    def get(self, ident):
        prop = self.property_name
        if self._criterion: # pragma: no cover
            mapper = self._only_full_mapper_zero("get")
            # Don't use getattr/hasattr to check public existence, since this
            # might misinterpret a bug (AttributeError raised by some code in
            # property implementation) as missing attribute and cause all
            # private data going to public.
            if prop in dir(mapper.class_):
                crit = getattr(mapper.class_, prop)
                if crit is not None:
                    if not isinstance(crit, ClauseElement):
                        # This simplest safe way to make bare boolean column
                        # accepted as expression.
                        crit = cast(crit, Boolean)
                    if crit!=self._criterion:
                        # We can't verify that criterion is from our private()
                        # call.  Check from DB instead of looking in identity
                        # map.
                        assert False # XXX temporal to verify it's used
                        return Query.get(self.populate_existing(), ident)
            assert False # XXX temporal to verify it's used
        obj = Query.get(self, ident)
        if obj is not None and (prop not in dir(obj) or getattr(obj, prop)):
            return obj

    def __iter__(self):
        return Query.__iter__(self.private())

    def from_self(self, *ent):
        # override from_self() to automatically apply
        # the criterion too.   this works with count() and
        # others.
        return Query.from_self(self.private(), *ent)

    def with_entities(self, *entities):
        return Query.with_entities(self.private(), *entities)

    def count(self):
        # Without it it works with slow implementation of count(), while
        # we often use a faster one from older version.
        return Query.count(self.private())

    def slice(self, start, stop):
        return Query.slice(self.private(), start, stop)

    def limit(self, limit):
        return Query.limit(self.private(), limit)

    def offset(self, offset):
        return Query.offset(self.private(), offset)

    def _entity_criterion(self, entity):
        #if hasattr(entity, "property"):
        #    entity = entity.property.mapper
        if hasattr(entity, 'parententity'):
            # XXX is this used?
            entity = entity.parententity
        try:
            cls = _class_to_mapper(entity).class_
        except AttributeError:
            # XXX For tables, table columns
            #pass
            raise # XXX temporal, to verify it's used
        else:
            alias = entity if isinstance(entity, AliasedClass) else cls
            prop = self.property_name
            if prop in dir(cls):
                crit = getattr(alias, prop)
                if crit is not None:
                    if not isinstance(crit, ClauseElement):
                        # This simplest safe way to make bare boolean column
                        # accepted as expression.
                        crit = cast(crit, Boolean)
                    return crit
        return None

    def _add_entity_criterion(self, entity):
        crit = self._entity_criterion(entity)
        if crit is not None:
            return self.filter(crit)
        return self

    def private(self):
        if self._limit is not None or self._offset is not None \
                or self._statement is not None:
            # Conditions must be added just before setting LIMIT and OFFSET
            # Calling it with statement means from_statement was used: either
            # manually (it's your problem) or by load_scalar_attributes (no
            # need in filtering here).
            return self
        query = self
        for query_entity in self._entities:
            for entity in query_entity.entities:
                query = query._add_entity_criterion(entity)
        for entity in self._join_entities:
            query = query._add_entity_criterion(entity)
        return query

    def _add_eager_onclause(self, obj, selectable, crit):
        # be careful! should return anything only if the criterion has been
        # applied already
        if isinstance(obj, Join):
            if obj is selectable or obj.right == selectable:
                obj = obj._clone()
                obj.onclause = obj.onclause & crit
                return obj
            left = self._add_eager_onclause(obj.left, selectable, crit)
            if left is not None:
                obj = obj._clone()
                obj.left = left
                return obj
            right = self._add_eager_onclause(obj.right, selectable, crit)
            if right is not None:
                obj = obj._clone()
                obj.right = right
                return obj
        if isinstance(obj, FromGrouping):
            # XXX tests required!
            element = self._add_eager_onclause(obj.element, selectable, crit)
            if element is not None:
                obj = obj._clone()
                obj.element = element
                return obj
        return None

    def _add_eager_criterion(self, context, statement):
        for attr, value in context.attributes.items():
            if type(attr) is tuple and attr[0] == 'eager_row_processor':
                alias = value.aliased_class
                crit = self._entity_criterion(alias)
                if crit is not None:
                    # add criterion to join "on" clause because if we add it
                    # to where clause we filter out objects with related 
                    # unpublished and only unpublished items
                    statement = statement._clone()
                    new_from_obj = []
                    for obj in statement._from_obj:
                        selectable = alias._aliased_insp.selectable
                        new_obj = self._add_eager_onclause(obj, selectable, crit)
                        obj = new_obj if new_obj is not None else obj
                        new_from_obj.append(obj)
                    statement._from_obj = type(statement._from_obj)(new_from_obj)
        return statement

    def _simple_statement(self, context):
        # fixing joined load
        # XXX is this solution correct?
        statement = Query._simple_statement(self, context)
        return self._add_eager_criterion(context, statement)

    def _compound_eager_statement(self, context):
        # fixing joined load
        # XXX not tested! What is condition when this method is called
        statement = Query._compound_eager_statement(self, context)
        return self._add_eager_criterion(context, statement)



# -*- coding: utf-8 -*-
from webob.exc import status_map, HTTPMethodNotAllowed
from .core import cases
from . import filters

__all__ = ['redirect_to', 'Rule']

def redirect_to(endpoint, _code=303, qs=None, **kwargs):
    def handle(env, data):
        url = env.root.build_url(endpoint, **kwargs)
        if qs is not None:
            url = url.qs_set(qs)
        raise status_map[_code](location=str(url.with_host()))
    return handle


def Rule(path, handler, method=None, name=None, convs=None):
    # werkzeug-style Rule
    if name is None:
        name = handler.__name__
    h = filters.match(path, name, convs=convs)
    if method is not None:
        h = h | cases(filters.method(method),
                      HTTPMethodNotAllowed)
    return h | handler




# -*- coding: utf-8 -*-
'''
Versioned storage, a classes for `env` and `data` objects.
'''

class StorageFrame(object):

    '''A single frame in the storage'''

    def __init__(self, _parent_storage=None, **kwargs):
        self._parent_storage = _parent_storage
        self.__dict__.update(kwargs)

    def as_dict(self):
        d = dict(self._parent_storage.as_dict() if self._parent_storage else {},
                 **self.__dict__)
        # XXX delete _parent_storage?
        del d['_parent_storage']
        if '_root_storage' in d:
            del d['_root_storage']
        return d

class VersionedStorage(object):
    '''Storage implements state managing interface, allowing to safely set
    attributes for `env` and `data` objects.

    Safity means that state of the storage is rolled back every time routing
    case returns ContinueRoute signal (`None`).

    Be very accurate defining methods or properties for storage, choose correct
    method or property type depending on what do you want to achieve.

    Regular methods will hold the state of storage frame they are added to.
    If you want to have an access to actual value, use storage property and
    method decorators.'''

    def __init__(self, cls=StorageFrame, *args, **kwargs):
        kwargs['_root_storage'] = self
        self._storage = cls(*args, **kwargs)

    def _push(self, **kwargs):
        self._storage = StorageFrame(_parent_storage=self._storage, **kwargs)
        return self._storage

    def _pop(self):
        self._storage = self._storage._parent_storage

    def __getattr__(self, name):
        frame = self._storage
        while frame:
            try:
                return getattr(frame, name)
            except AttributeError:
                frame = frame._parent_storage
        #raise AttributeError(name)
        raise AttributeError("{} has no attribute {}".format(
                             self.__class__.__name__, name))

    def __setattr__(self, name, value):
        if name == '_storage':
            self.__dict__[name] = value
        else:
            setattr(self._storage, name, value)

    def __delattr__(self, name):
        delattr(self._storage, name)

    def as_dict(self):
        '''Returns attributes of storage as dict'''
        return self._storage.as_dict()


class storage_property_base(object):

    def __init__(self, method, name=None):
        self.method = method
        self.name = name or method.__name__
        self.__doc__ = method.__doc__


class storage_property(storage_property_base):
    '''Turns decorated method into storage property (method is called with
       VersionedStorage as self).'''

    def __get__(self, inst, cls):
        if inst is None:
            return self
        return self.method(inst._root_storage)


class storage_cached_property(storage_property_base):
    '''Turns decorated method into storage cached property 
       (method is called only once with VersionedStorage as self).'''

    def __get__(self, inst, cls):
        if inst is None:
            return self
        result = self.method(inst._root_storage)
        setattr(inst, self.name, result)
        return result

def storage_method(func):
    '''Calls decorated method with VersionedStorage as self'''
    def wrap(self, *args, **kwargs):
        return func(self._root_storage, *args, **kwargs)
    return wrap



# -*- coding: utf-8 -*-

from . import convs, widgets, fields
from iktomi.utils.i18n import N_


class PasswordConv(convs.Char):

    error_mismatch = N_('password and confirm mismatch')
    error_required = N_('password required')

    def from_python(self, value):
        return dict([(field.name, None) for field in self.field.fields])

    def get_initial(self):
        return ''

    def to_python(self, value):
        etalon = value[list(value)[0]]
        for field in self.field.fields:
            self.assert_(value[field.name] == etalon,
                         self.error_mismatch)
        if self.required:
            self.assert_(etalon not in (None, ''), self.error_required)
        elif etalon in (None, ''):
            return None
        return etalon


def PasswordSet(name='password',
                 min_length=3, max_length=200, required=False,
                 password_label=None, confirm_label='confirm', filters=(),
                 **kwargs):
        # class implementation has problem with Fieldset copying:
        # it requires to save all kwargs in object's __dict__
        char = convs.Char(convs.length(min_length, max_length), *filters,
                          **dict(required=required))
        items = (('pass', password_label), ('conf', confirm_label))
        kwargs['fields'] = [fields.Field(subfieldname,
                                         conv=char,
                                         label=label,
                                         widget=widgets.PasswordInput)
                            for subfieldname, label in items]
        kwargs.setdefault('conv', PasswordConv(required=required))
        kwargs.setdefault('widget', widgets.FieldSetWidget(
            template='widgets/fieldset-line'))
        return fields.FieldSet(name, get_initial=lambda: '', **kwargs)




# -*- coding: utf-8 -*-
'''
Module containing some predefined form converters - objects designed to
convert and validate form field's data.
'''

import six
import re
from ..utils import weakproxy, replace_nontext
from datetime import datetime
from collections import OrderedDict
from ..utils.dt import strftime
from ..utils.deprecation import deprecated
try:
    from ..utils.html import Cleaner
    from lxml import html
    from lxml.etree import XMLSyntaxError
except ImportError: # pragma: no cover
    # lxml is required for Html conv, therefore you can use forms without this
    # functionality
    Cleaner = None

from iktomi.utils import cached_property
from iktomi.utils.i18n import N_, M_
from sqlalchemy.ext.associationproxy import _AssociationCollection


_all2 = set(vars())



class ValidationError(Exception):
    '''
    Error raised from inside of `Converter.to_python` or validator function.

    :param unicode message: error message for current validating
        field, for most cases.
    :param dict by_field: contains {field-name: error message} pairs.
    :param dict format_args: used to format error message.
    '''

    default_message = N_('Something is wrong')

    def __init__(self, message=None, by_field=None, format_args=None):
        if not (message or by_field):
            message = self.default_message
        self.message = message
        self.by_field = by_field or {}
        self.format_args = format_args or {}

    def translate(self, env, message):
        format_args = self.format_args
        if isinstance(message, M_):
            trans = env.ngettext(message.single,
                                 message.plural,
                                 message.count)
            format_args = dict(format_args,
                               **message.format_args)
        else:
            trans = env.gettext(message)
        return trans % format_args

    def fill_errors(self, field):
        form = field.form
        if self.message is not None:
            form.errors[field.input_name] = self.translate(form.env, self.message)
        for name, message in self.by_field.items():
            if name.startswith('.'):
                nm, f = name.lstrip('.'), field
                for i in range(len(name) - len(nm) - 1):
                    f = f.parent
                rel_field = f.get_field(nm)
                name = rel_field.input_name
            else:
                rel_field = form.get_field(name)
            form.errors[name] = self.translate(form.env, message)

    def __repr__(self):
        return "{}({!r}, {!r})".format(self.__class__, self.message,
                                       self.by_field)


class Converter(object):
    '''
    :meth:`accept` method takes value from field
    and converts it to python type.

    Subclasses must redefine :meth:`to_python` method to make their
    own convertation and validation.

    :meth:`from_python` method takes value as python
    object and converts it to string or something
    else widget can display.

    Accepts a list of validators as **\*args**.
    '''

    # obsolete parameters from previous versions
    _obsolete = frozenset(['max_length', 'min_length', 'null', 'min', 'max',
                           'multiple', 'initial'])
    #: Flag of whether perform require check after :meth:`to_python` method or not.
    #: The resulting value is checked to be non-empty (`[]`, `None`).
    required = False
    multiple = False

    #: An ordered list of validator functions. Are passed as position args
    #: to the converter::
    #:
    #:     Int(validator1, validator2)
    #:
    #: When a converter is copied, new validators are added to existing ones.
    validators = ()

    #: Values are not accepted by Required validator
    error_required = N_('required field')

    def __init__(self, *args, **kwargs):
        if self._obsolete & set(kwargs):
            raise TypeError(
                    'Obsolete parameters are used: {}'.format(
                                list(self._obsolete & set(kwargs))))
        self.field = weakproxy(kwargs.get('field'))
        self._init_kwargs = kwargs
        self.__dict__.update(kwargs)
        self.validators = tuple(self.validators) + args

    @property
    def env(self):
        '''A shortcut for `form.env`'''
        return self.field.env

    def _is_empty(self, value):
        return value in ('', [], {}, None)

    def accept(self, value, silent=False):
        '''
        Accepts a value from the form, calls :meth:`to_python` method,
        checks `required` condition, applies filters and validators,
        catches ValidationError.

        :param value: a value to be accepted
        :param silent=False: write errors to `form.errors` or not
        '''
        try:
            value = self.to_python(value)
            for v in self.validators:
                value = v(self, value)

            if self.required and self._is_empty(value):
                raise ValidationError(self.error_required)
        except ValidationError as e:
            if not silent:
                e.fill_errors(self.field)
            #NOTE: by default value for field is in python_data,
            #      but this is not true for FieldList where data
            #      is dynamic, so we set value to None for absent value.
            value = self._existing_value
        return value

    def to_python(self, value):
        """
        Converts value and validates it.
        Custom converters should override this
        """
        if value == '':
            return None # XXX is this right?
        return value

    def from_python(self, value):
        """
        Serializes value.
        Custom converters should override this
        """
        if value is None:
            value = ''
        return value

    def __call__(self, *args, **kwargs):
        '''
        Creates current object's copy with extra constructor arguments
        (including validators) passed.
        '''
        kwargs = dict(self._init_kwargs, **kwargs)
        kwargs.setdefault('field', self.field)
        validators = kwargs.pop('validators', self.validators)
        validators = tuple(validators) + args
        return self.__class__(*validators, **kwargs)

    def assert_(self, expression, msg):
        'Shortcut for assertions of certain type'
        if not expression:
            raise ValidationError(msg)

    @property
    def _existing_value(self):
        if self.field is not None:
            return self.field.parent.python_data.get(self.field.name)
        return [] if self.multiple else None

    def __repr__(self):
        args = ', '.join([k+'='+repr(v) for
                          k, v in self._init_kwargs.items()
                          if k!='parent'])
        return '{}({})'.format(self.__class__.__name__, args)


class validator(object):
    'Function decorator'
    def __init__(self, message):
        self.message = message
    def __call__(self, func):
        def wrapper(conv, value):
            if not func(conv, value):
                raise ValidationError(self.message)
            return value
        return wrapper

# Some useful validators

class length(object):
    'String length constraint'

    func_name = 'check_length' # XXX for backward compatibility

    def __init__(self, min_length, max_length):
        self.min_length = min_length
        self.max_length = max_length

        self.format_args = dict(min=min_length, max=max_length)

        if min_length == max_length:
            self.message = M_(u'length of value must be exactly %(max)d symbol',
                              u'length of value must be exactly %(max)d symbols',
                              count_field="max",
                              format_args=self.format_args)
        else:
            self.message = N_('length should be between %(min)d and %(max)d symbols')

    def __call__(self, conv, value):
        if value and not (self.min_length <= len(value) <= self.max_length):
            raise ValidationError(self.message, format_args=self.format_args)
        return value


@deprecated('Use length(min, max) instead.')
def limit(min_length, max_length):
    return length(min_length, max_length) # pragma: no cover


def between(min_value, max_value):
    'Numerical values limit'
    message = N_('value should be between %(min)d and %(max)d') % \
                    dict(min=min_value, max=max_value)

    @validator(message)
    def wrapper(conv, value):
        if value is None:
            # it meens that this value is not required
            return True
        if value < min_value:
            return False
        if value > max_value:
            return False
        return True
    return wrapper

@deprecated('Use between(min, max) instead.')
def num_limit(min_value, max_value):
    return between(min_value, max_value) # pragma: no cover


class CharBased(Converter):

    nontext_replacement = u'\uFFFD' # Set None to disable and empty string to
                                    # remove.
                                    # Default value is u"�"
    #: Whether strip value before convertation or not
    strip = True

    def clean_value(self, value):
        '''
        Additional clean action to preprocess value before :meth:`to_python`
        method.

        Subclasses may define own clean_value method to allow additional clean
        actions like html cleanup, etc.
        '''
        # We have to clean before checking min/max length. It's done in
        # separate method to allow additional clean action in subclasses.
        if self.nontext_replacement is not None:
            value = replace_nontext(value, self.nontext_replacement)
        if self.strip:
            value = value.strip()
        return value


class Char(CharBased):

    '''Converts and validates strings'''

    #: Regexp to match input string
    regex = None
    #: Error message for the case self.regex does not match
    error_regex = N_('field should match %(regex)s')
    #: Whether strip value before convertation or not
    strip = True

    @property
    def max_length(self):
        length_validators = [x for x in self.validators
                             if isinstance(x, length)]
        if length_validators:
            return min([x.max_length for x in length_validators])
        return None

    def to_python(self, value):
        # converting
        value = self.clean_value(value)
        if value and self.regex:
            regex = self.regex
            if isinstance(regex, six.string_types):
                regex = re.compile(self.regex, re.U)
            if not regex.match(value):
                error = self.error_regex % {'regex': self.regex}
                raise ValidationError(error)
        return value

    def from_python(self, value):
        if value is None:
            return ''
        if six.PY3 and isinstance(value, bytes):
            raise TypeError() # pragma: no cover, safety check
        return six.text_type(value)


class Int(Converter):
    """
    Converts digital sequences to `int`
    """

    #: Error message for the case value can not be converted to int
    error_notvalid = N_('it is not valid integer')

    def to_python(self, value):
        if value == '':
            return None
        try:
            value = int(value)
        except ValueError:
            raise ValidationError(self.error_notvalid)
        return value

    def from_python(self, value):
        if value is None:
            return ''
        return six.text_type(value)


class Bool(Converter):
    """
    Converts to `True`/`False`
    """
    required = False

    def to_python(self, value):
        return bool(value)

    def from_python(self, value):
        if value:
            return 'checked'
        return ''


class EnumChoice(Converter):
    '''
    In addition to Converter interface it must provide
    :meth:`options` and :meth:`get_label` methods.
    '''

    #: converter for value, before it is tested to be in a set of acceptable
    #: values
    conv = Char()

    #: acceptable choices list::
    #:
    #:     EnumChoice(choices=[(python_value, label), ...])
    choices = ()
    error_required = N_('you must select a value')

    def from_python(self, value):
        conv = self.conv
        return conv.from_python(value)

    def to_python(self, value):
        value = self.conv.accept(value, silent=True)
        if value not in dict(self.choices):
            return None
        return value

    def options(self):
        '''
        Yields `(raw_value, label)` pairs for all acceptable choices.
        '''
        conv = self.conv
        for python_value, label in self.choices:
            yield conv.from_python(python_value), label


class BaseDatetime(CharBased):
    '''
    A base class for `Datetime`, `Date` and `Time` converters.
    '''

    #: format used to convert from string by `strptime`
    #: and to convert to string by `strftime`.
    format = None
    #: format used in error message. By default, generated automatically
    #: based `format` and `replacements` attributes
    readable_format = None
    replacements = (('%H', 'HH'), ('%M', 'MM'), ('%d', 'DD'),
                    ('%m', 'MM'), ('%Y', 'YYYY'))
    #: error message for wrong format.
    error_wrong_format = N_('Wrong format (%(readable_format)s)')
    nontext_replacement = ''

    def __init__(self, *args, **kwargs):
        if not 'readable_format' in kwargs or 'format' in kwargs:
            replacements = self.replacements # XXX make language-dependent
            fmt = kwargs.get('format', self.format)
            for repl in replacements:
                fmt = fmt.replace(*repl)
            kwargs['readable_format'] = fmt
        Converter.__init__(self, *args, **kwargs)

    def from_python(self, value):
        if value is None:
            return ''
        # carefull to years before 1900
        return strftime(value, self.format)

    def to_python(self, value):
        value = self.clean_value(value)
        if not value:
            return None
        try:
            return self.convert_datetime(value)
        except ValueError:
            format_args = dict(readable_format=self.readable_format)
            raise ValidationError(self.error_wrong_format,
                                  format_args=format_args)


class Datetime(BaseDatetime):
    '''
    Subclass of `BaseDatetime` for `datetime.datetime`
    '''

    format = '%d.%m.%Y, %H:%M'

    def convert_datetime(self, value):
        return datetime.strptime(value, self.format)


class Date(BaseDatetime):
    '''
    Subclass of `BaseDatetime` for `datetime.date`
    '''

    format = '%d.%m.%Y'

    def convert_datetime(self, value):
        return datetime.strptime(value, self.format).date()


class Time(BaseDatetime):
    '''
    Subclass of `BaseDatetime` for `datetime.time`
    '''

    format = '%H:%M'

    def from_python(self, value):
        if value is None:
            return ''
        # we don't care about year in time converter, so use native strftime
        return value.strftime(self.format)

    def convert_datetime(self, value):
        return datetime.strptime(value, self.format).time()


class DisplayOnly(Converter):
    """
    Does nothing, always returns a value of field before validation
    """

    def from_python(self, value):
        return value

    def to_python(self, value):
        return self._existing_value


class SplitDateTime(Converter):
    '''
    Converter for FieldSet, allowing get a single `datetime.datetime`
    object from two fields: date and time.
    '''

    def from_python(self, value):
        if value is None:
            return {'date': None, 'time': None}
        else:
            return {'date': value.date(), 'time': value.time()}

    def to_python(self, value):
        if value['date'] is None or value['time'] is None:
            return None
        res = datetime.combine(value['date'], value['time'])
        return res


class Html(Char):
    '''
    Converter for flexible cleanup of HTML document fragments.
    A subclass of :class:`Char<iktomi.forms.convs.Char>`.

    Uses :class:`utils.html.Sanitizer<iktomi.utils.html.Sanitizer>`
    instance to sanitize input HTML.

    Construtor collects from given kwargs all of
    :class:`Sanitizer<iktomi.utils.html.Sanitizer>`
    options and passes them into Sanitizer's constructor.

    For list properties there is :meth:`add_*` interface::

        Html(add_allowed_elements=['span'], add_dom_callbacks=[myfunc])
    '''

    #: A list of allowed HTML elements
    allowed_elements = frozenset(('a', 'p', 'br', 'li', 'ul', 'ol', 'hr', 'u',
                                  'i', 'b', 'blockquote', 'sub', 'sup'))
    #: A list of allowed HTML attributes
    allowed_attributes = frozenset(('href', 'src', 'alt', 'title', 'class',
                                    'rel'))
    #: A list of tags to be dropped if they are empty
    drop_empty_tags = frozenset(('p', 'a', 'u', 'i', 'b', 'sub', 'sup'))
    allowed_protocols = frozenset(['ftp', 'http', 'https', 'mailto',
                                   'tel', 'webcal', 'callto'])
    #: A dict containing an element name as a key and class test as value.
    #: Test can be a set of strings (for simple strict match) or callable
    #: accepting class name and returning True/False.
    #: Set to `None` to allow all classes.
    allowed_classes = {}
    #: A list of callbacks applied to DOM tree before it is
    #: rendered back to HTML.
    dom_callbacks = []
    wrap_inline_tags = None
    split_paragraphs_by_br = True
    # Tags to wrap in paragraphs on top
    tags_to_wrap = ['b', 'big', 'i', 'small', 'tt',
                    'abbr', 'acronym', 'cite', 'code',
                    'dfn', 'em', 'kbd', 'strong', 'samp',
                    'var', 'a', 'bdo', 'br', 'map', 'object',
                    'q', 'span', 'sub', 'sup']
    #: Function returning object marked safe for template engine.
    #: For example, `jinja2.Markup` object.
    Markup = lambda s, x: x
    #: A Cleaner class. Be default, 'iktomi.utils.html.Cleaner' is used.
    #: It is a `lxml.html.clean.Cleaner` subclass
    Cleaner = Cleaner
    class Nothing: pass

    PROPERTIES = ['allowed_elements', 'allowed_attributes', 'allowed_protocols',
                  'allowed_classes', 'dom_callbacks', 'drop_empty_tags',
                  'wrap_inline_tags', 'split_paragraphs_by_br']

    LIST_PROPERTIES = ['allowed_elements', 'allowed_attributes',
                       'allowed_protocols', 'dom_callbacks',
                       'drop_empty_tags', 'tags_to_wrap']

    @classmethod
    def _load_arg(cls, opt):
        assert hasattr(cls, opt)
        # XXX very complicated merges for rare cases
        #     not sure if they are necessary
        if opt in cls.__dict__:
            result = getattr(cls, opt)
        else:
            for base in cls.__bases__:
                if hasattr(base, '_load_arg'):
                    result = base._load_arg(opt)
                    if not result is cls.Nothing:
                        break
            else:
                result = cls.Nothing

        if 'add_' + opt in cls.__dict__:
            add = getattr(cls, 'add_' + opt)
            if result is cls.Nothing:
                result = add
            else:
                result = set(result)
                result.update(add)
        return result

    def __init__(self, *args, **kwargs):
        assert self.Cleaner is not None, \
                'Install lxml or implement your own html cleaner'
        for opt in self.PROPERTIES:
            if not opt in kwargs:
               result = self._load_arg(opt)
               if not result is self.Nothing:
                   kwargs[opt] = result

        for opt in self.LIST_PROPERTIES:
            add_key = 'add_' + opt
            if add_key in kwargs:
                opt_value = kwargs.get(opt, [])
                # XXX sometimes set must be ordered
                kwargs[opt] = set(opt_value)
                kwargs[opt].update(kwargs.pop(add_key))
        Char.__init__(self, *args, **kwargs)

    def clean_value(self, value):
        value = Char.clean_value(self, value)
        try:
            doc = html.fragment_fromstring(value, create_parent=True)
        except XMLSyntaxError: # pragma: no cover. XXX: seems like this exception is
                               # unreachable with create_parent=True,
                               # maybe it should be removed?
            raise ValidationError(N_(u'Error parsing HTML'))
        self.cleaner(doc)
        clean = html.tostring(doc, encoding='utf-8').decode('utf-8')
        clean = clean.split('>', 1)[1].rsplit('<', 1)[0]
        return self.Markup(clean)

    @cached_property
    def cleaner(self):
        return self.Cleaner(allow_tags=self.allowed_elements,
                            safe_attrs=self.allowed_attributes,
                            allow_classes=self.allowed_classes,
                            allowed_protocols=self.allowed_protocols,
                            drop_empty_tags=self.drop_empty_tags,
                            dom_callbacks=self.dom_callbacks,
                            wrap_inline_tags=self.wrap_inline_tags,
                            split_paragraphs_by_br=self.split_paragraphs_by_br,
                            tags_to_wrap=self.tags_to_wrap,
                            )


class List(Converter):
    '''
    Converter for FieldList'''

    _obsolete = Converter._obsolete | set(['filter'])

    def from_python(self, value):
        if value is None:
            value = []
        result = OrderedDict()
        for index, item in enumerate(value):
            result[str(index+1)] = item
        return result

    def to_python(self, value):
        return list(value.values())


class ListOf(Converter):
    '''
    Converter for scalar Fields, applies nested converter to each value
    of the field (i.e. for each value from MultiDict) and returns a list of
    resulting values.

    Usage::

        ListOf(Converter(), *validators, **kwargs)
    '''

    multiple = True

    def __init__(self, *args, **kwargs):
        if args and (isinstance(args[0], Converter) or \
                     (isinstance(args[0], type) and \
                      issubclass(args[0], Converter))):
            conv = args[0]
            args = args[1:]
        else:
            conv = kwargs['conv']
        if 'field' in kwargs:
            conv = conv(field=kwargs['field'])
        kwargs['conv'] = conv
        Converter.__init__(self, *args, **kwargs)

    def to_python(self, value):
        result = []
        for val in value or []:
            val = self.conv.accept(val)
            if val is not None:
                # XXX is it right to ignore None?
                result.append(val)
        return result

    def from_python(self, value):
        return [self.conv.from_python(item) for item in value or []]


class FieldBlockConv(Converter):

    @property
    def _existing_value(self):
        if self.field is not None:
            return self.field.python_data
        return {} # XXX


class SimpleFile(Converter):
    '''
    Simpliest converter for files. Returns a file from webob Request.POST
    "as is".
    '''

    def _is_empty(self, file):
        return file == u'' or file is None #XXX WEBOB ONLY !!!

    def to_python(self, file):
        if not self._is_empty(file):
            return file

    def from_python(self, value):
        return None


class Email(Char):

    regex = re.compile(
        # dot-atom
        r"(^[-!#$%&'*+/=?^_`{}|~0-9A-Z]+(\.[-!#$%&'*+/=?^_`{}|~0-9A-Z]+)*"
        # Although quoted variant is allowed by spec it's actually not used
        # except by geeks that are looking for problems. But the characters
        # allowed in quoted string are not safe for HTML and XML, so quoted
        # e-mail can't be expressed in such formats directly which is quite
        # common. We prefer to forbid such valid but unsafe e-mails to avoid
        # security problems. To allow quoted names disable non-text characters
        # replacement and uncomment the following lines of regexp:
        ## quoted-string
        #r'|^"([\001-\010\013\014\016-\037!#-\[\]-\177]|'
        #    r'\\[\001-011\013\014\016-\177])*"'
        r')@(?:[A-Z0-9-]+\.)+[A-Z]{2,6}$', re.IGNORECASE)
    error_regex = N_('incorrect e-mail address')


class ModelDictConv(Converter):
    '''Converts a dictionary to object of `model` class with the same fields.
    It is designed for use in FieldSet'''

    model = None

    def from_python(self, value):
        if value is None:
            # Field set can be optional
            return {}
        result = {}
        field_names = sum([x.field_names for x in self.field.fields], [])
        for field_name in field_names:
            attr = getattr(value, field_name)
            if issubclass(attr.__class__, _AssociationCollection):
                attr = attr.copy()
            result[field_name] = attr
        return result

    def to_python(self, value):
        obj = self.model()
        field_names = sum([x.field_names for x in self.field.fields], [])
        for field_name in field_names:
            field = self.field.get_field(field_name)
            if field.writable:
                setattr(obj, field_name, value[field_name])
        return obj

    @property
    def _existing_value(self):
        if self.field is not None:
            pd = self.field.parent.python_data
            if self.field.name in pd:
                return pd[self.field.name]
            # Return blank self.model instance as initial/default value 
            # if one does not exist
            return self.model()
        return None


class OptionLabel(six.text_type):

    published = False


class ModelChoice(EnumChoice):

    condition = None
    conv = Int(required=False)
    title_field = 'title'

    def __init__(self, *args, **kwargs):
        EnumChoice.__init__(self, *args, **kwargs)
        self.conv = self.conv(field=self.field)

    @property
    def query(self):
        query = self.env.db.query(self.model)
        if isinstance(self.condition, dict):
            query = query.filter_by(**self.condition)
        elif self.condition is not None:
            query = query.filter(self.condition)
        return query

    def from_python(self, value):
        if value is not None:
            return self.conv.from_python(value.id)
        else:
            return ''

    def to_python(self, value):
        try:
            value = self.conv.to_python(value)
        except ValidationError:
            return None
        else:
            if value is not None:
                return self.query.filter_by(id=value).first()

    def get_object_label(self, obj):
        label = OptionLabel(getattr(obj, self.title_field))
        try:
            label.published = obj.publish
        except AttributeError:
            pass
        return label

    def options(self):
        for obj in self.query.all():
            yield self.conv.from_python(obj.id), self.get_object_label(obj)




# Expose all variables defined after imports
__all__ = [x for x
           in set(vars()) - _all2
           if not x.startswith('_')]
del _all2




# -*- coding: utf-8 -*-
import six
import sys
import os
import datetime

__all__ = ['manage']


class CommandNotFound(AttributeError): pass
class ConverterError(Exception): pass


def manage(commands, argv=None, delim=':'):
    '''
    Parses argv and runs neccessary command. Is to be used in manage.py file.

    Accept a dict with digest name as keys and instances of
    :class:`Cli<iktomi.management.commands.Cli>`
    objects as values.

    The format of command is the following::

        ./manage.py digest_name:command_name[ arg1[ arg2[...]]][ --key1=kwarg1[...]]

    where command_name is a part of digest instance method name, args and kwargs
    are passed to the method. For details, see
    :class:`Cli<iktomi.management.commands.Cli>` docs.
    '''

    commands = {(k.decode('utf-8') if isinstance(k, six.binary_type) else k): v
                for k, v in commands.items()}

    # Default django autocompletion script is registered to manage.py
    # We use the same name for this script and it seems to be ok
    # to implement the same interface
    def perform_auto_complete(commands):
        from .lazy import LazyCli
        cwords = os.environ['COMP_WORDS'].split()[1:]
        cword = int(os.environ['COMP_CWORD'])

        try:
            curr = cwords[cword - 1]
        except IndexError:
            curr = ''

        suggest = []
        if len(cwords) > 1 and cwords[0] in commands.keys():
            value = commands[cwords[0]]
            if isinstance(value, LazyCli):
                value = value.get_digest()

            for cmd_name, _ in value.get_funcs():
                cmd_name = cmd_name[8:]
                suggest.append(cmd_name)
            if curr == ":":
                curr = ''
        else:
            suggest += list(commands.keys()) + [x+":" for x in commands.keys()]
        suggest.sort()
        output = u" ".join(filter(lambda x: x.startswith(curr), suggest))
        sys.stdout.write(output)

    auto_complete = 'IKTOMI_AUTO_COMPLETE' in os.environ or \
                    'DJANGO_AUTO_COMPLETE' in os.environ
    if auto_complete:
        perform_auto_complete(commands)
        sys.exit(0)

    argv = sys.argv if argv is None else argv
    if len(argv) > 1:
        cmd_name = argv[1]
        raw_args = argv[2:]
        args, kwargs = [], {}
        # parsing params
        for item in raw_args:
            if item.startswith('--'):
                splited = item[2:].split('=', 1)
                if len(splited) == 2:
                    k,v = splited
                elif len(splited) == 1:
                    k,v = splited[0], True
                kwargs[k] = v
            else:
                args.append(item)

        # trying to get command instance
        if delim in cmd_name:
            digest_name, command = cmd_name.split(delim)
        else:
            digest_name = cmd_name
            command = None
        try:
            digest = commands[digest_name]
        except KeyError:
            _command_list(commands)
            sys.exit('ERROR: Command "{}" not found'.format(digest_name))
        try:
            if command is None:
                if isinstance(digest, Cli):
                    help_ = digest.description(argv[0], digest_name)
                    sys.stdout.write(help_)
                    sys.exit('ERROR: "{}" command digest requires command name'\
                                .format(digest_name))
                digest(*args, **kwargs)
            else:
                digest(command, *args, **kwargs)
        except CommandNotFound:
            help_ = digest.description(argv[0], digest_name)
            sys.stdout.write(help_)
            sys.exit('ERROR: Command "{}:{}" not found'.format(digest_name, command))
    else:
        _command_list(commands)
        sys.exit('Please provide any command')

def _command_list(commands):
    sys.stdout.write(u'Commands:\n')
    for k in commands.keys():
        sys.stdout.write(k)
        sys.stdout.write(u'\n')


class Cli(object):
    'Base class for all command digests'

    def _fix_docstring(self, doc, indent=0):
        doc = doc.lstrip('\n').rstrip()
        doc = doc.split('\n\n')[0].rstrip(':')
        doc =  '\n'.join('\t'*indent + x.lstrip()
                         for x in doc.splitlines())
        doc += '\n'
        return doc

    def get_funcs(self):
        return [(attr, getattr(self, attr))
                 for attr in dir(self)
                 if attr.startswith('command_')]

    def description(self, argv0='manage.py', command=None):
        '''Description outputed to console'''
        command = command or self.__class__.__name__.lower()
        import inspect
        _help = u''
        _help += u'{}\n'.format(command)
        if self.__doc__:
            _help += self._fix_docstring(self.__doc__) +'\n'
        else:
            _help += u'{}\n'.format(command)

        funcs = self.get_funcs()
        funcs.sort(key=lambda x: six.get_function_code(x[1]).co_firstlineno)

        for attr, func in funcs:
            func = getattr(self, attr)
            comm = attr.replace('command_', '', 1)
            if six.PY2:
                args = inspect.getargspec(func).args[1:]
            else:
                args = inspect.getfullargspec(func).args[1:]
            args = (' [' + '] ['.join(args) + ']') if args else ''

            _help += "\t{} {}:{}{}\n".format(
                            argv0, command, comm, args)

            if func.__doc__:
                _help += self._fix_docstring(func.__doc__, 2)
        return _help

    def __call__(self, command_name, *args, **kwargs):
        if command_name == 'help':
            sys.stdout.write(self.description())
        elif hasattr(self, 'command_'+command_name):
            try:
                getattr(self, 'command_'+command_name)(*args, **kwargs)
            except ConverterError as e:
                sys.stderr.write(u'One of the arguments for '
                                 u'command "{}" is wrong:\n'.format(command_name))
                sys.exit(e)
        else:
            raise CommandNotFound()


class argument(object):
    'Decorator. Helps to define desired argument types'

    def __init__(self, arg_id, *validators, **kwargs):
        assert type(arg_id) in (int, str), \
            'First argument of argument ({!r}) decorator must be "str" '\
            'or "int" type'.format(arg_id)
        self.arg_id = arg_id
        self.validators = validators
        self.required = kwargs.get('required', True)

    def __call__(self, func):
        def wrapper(*args, **kwargs):
            args = list(args)
            if isinstance(self.arg_id, int):
                try:
                    arg = args[self.arg_id]
                except IndexError:
                    raise ConverterError('Total positional args = {:d}, '
                        'but you apply converter for {:d} argument '
                        '(indexing starts from 0)'.format(len(args), self.arg_id))
                arg = self.convert(arg)
                args[self.arg_id] = arg
            elif isinstance(self.arg_id, str):
                arg = kwargs.get(self.arg_id)
                if arg:
                    arg = self.convert(arg)
                    kwargs[self.arg_id] = arg
                if not arg and self.required:
                    raise ConverterError('Keyword argument "{}" is required'\
                                            .format(self.arg_id))
            return func(*args, **kwargs)
        return wrapper

    def convert(self, arg):
        for v in self.validators:
            arg = v(arg)
        return arg

    @staticmethod
    def to_int(value):
        try:
            return int(value)
        except Exception:
            raise ConverterError('Cannot convert {!r} to int'.format(value))

    @staticmethod
    def to_date(value):
        try:
            return datetime.datetime.strptime(value, '%d/%m/%Y').date()
        except Exception:
            raise ConverterError('Cannot convert {!r} to date, please provide '
                                 'string in format "dd/mm/yyyy"'.format(value))



from iktomi.utils import cached_property
from .base import Cli

class LazyCli(Cli):
    '''
    Wrapper for creating lazy command digests.

    Sometimes it is not needed to import all of application parts to start
    a particular command. LazyCli allows you to define all imports in a 
    function called only on the command::

        @LazyCli
        def db_command():
            import admin
            from admin.environment import db_maker

            from models import initial
            from iktomi.cli import sqla
            return sqla.Sqla(db_maker, initial=initial.install)

        # ...

        def run(args=sys.argv):
            manage(dict(db=db_command, ), args)
    '''
    def __init__(self, func):
        self.get_digest = func

    @cached_property
    def digest(self):
        return self.get_digest()

    def description(self, *args, **kwargs):
        return self.digest.description(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        return self.digest(*args, **kwargs)



'''
Functions for application-level replication.

Therminology:

reflect::
    Find an object with the same identifier in the target database without
    changing or creating it.
replicate::
    Find or create new object with the same identifier in the target database
    and update it with data of the current object. Only SQLAlchemy attributes
    found in both source and target classes are copied. For objects found via
    relationships the following rules apply: private ones are replecated and
    references to independent objects are reflected.
'''

from weakref import WeakSet
from sqlalchemy.schema import Column
from sqlalchemy.util import duck_type_collection
from sqlalchemy.orm import object_session
from sqlalchemy.orm.util import identity_key
from sqlalchemy.orm.attributes import manager_of_class, QueryableAttribute
from sqlalchemy.orm.properties import ColumnProperty, RelationshipProperty
from sqlalchemy.orm.collections import collection_adapter
from sqlalchemy.orm.attributes import instance_state, instance_dict
from sqlalchemy.orm.interfaces import MANYTOMANY, MANYTOONE, ONETOMANY


_included = WeakSet()
_excluded = WeakSet()


def include(prop):
    '''Replicate property that is normally not replicated. Right now it's
    meaningful for one-to-many relations only.'''
    if isinstance(prop, QueryableAttribute):
        prop = prop.property
    assert isinstance(prop, (Column, ColumnProperty, RelationshipProperty))
    #assert isinstance(prop, RelationshipProperty)
    _included.add(prop)

def exclude(prop):
    '''Don't replicate property that is normally replicated: ordering column,
    many-to-one relation that is marked for replication from other side.'''
    if isinstance(prop, QueryableAttribute):
        prop = prop.property
    assert isinstance(prop, (Column, ColumnProperty, RelationshipProperty))
    _excluded.add(prop)
    if isinstance(prop, RelationshipProperty):
        # Also exclude columns that participate in this relationship
        for local in prop.local_columns:
            _excluded.add(local)


def reflect(source, model, cache=None):
    '''Finds an object of class `model` with the same identifier as the
    `source` object'''
    if source is None:
        return None
    if cache and source in cache:
        return cache[source]
    db = object_session(source)
    ident = identity_key(instance=source)[1]
    assert ident is not None
    return db.query(model).get(ident)


class _PrimaryKeyIsNull(BaseException):
    '''Used when setting relationship property to None if this causes setting
    not nullable primary key column to NULL. Such objects should be skipped
    from replicate_filter.'''


def replicate_relation(source, target, attr, target_attr, cache=None):
    if attr.property.cascade.delete_orphan:
        process_scalar = replicate_no_merge
        process_list = replicate_filter
    else:
        process_scalar = reflect
        process_list = reflect_filter
    value = getattr(source, attr.key)
    target_attr_model = target_attr.property.mapper.class_
    if attr.property.uselist:
        adapter = collection_adapter(value)
        if adapter:
            # XXX The magic passes below are adapted from logic in
            # CollectionAttributeImpl.set() method without proper
            # understanding.  The `elif` branch isn't even coverered by tests.
            if hasattr(value, '_sa_iterator'):
                value = value._sa_iterator()
            elif duck_type_collection(value) is dict:
                value = value.values()
        reflection = process_list(value, target_attr_model, cache=cache)
        impl = instance_state(target).get_impl(attr.key)
        impl.set(instance_state(target), instance_dict(target), reflection,
                 # XXX We either have to convert reflection back to original
                 # collection type or use this private parameter.
                 _adapt=False)
    else:
        reflection = process_scalar(value, target_attr_model, cache=cache)
        setattr(target, attr.key, reflection)
        if (reflection is None and
                attr.property.direction is MANYTOONE and
                any(col.primary_key and not col.nullable
                    for col in attr.property.local_columns)):
            raise _PrimaryKeyIsNull()


def is_relation_replicatable(attr):
    if attr.property in _included:
        return True
    elif attr.property in _excluded:
        return False
    elif attr.property.viewonly:
        return False
    elif attr.property.cascade.delete_orphan:
        # Private, replicate
        return True
    elif attr.property.direction is MANYTOMANY:
        # Many-to-many. Usualy one side is short list and other is long or
        # absent. Reflect if not dynamic, other cases should be excluded
        # manually.
        assert attr.property.lazy in (True, False, 'dynamic', 'select')
        return attr.property.lazy!='dynamic'
    elif attr.property.direction is MANYTOONE:
        # Many-to-one and one-to-one with FK pointing from from this side to
        # other.
        return True
    else:
        assert attr.property.direction is ONETOMANY
        return False


def _column_property_in_registry(prop, registry):
    if prop in registry:
        return True
    elif len(prop.columns)==1:
        # Column() is translated to ColumnProperty with single column
        return prop.columns[0] in registry
    else:
        return False


def replicate_attributes(source, target, cache=None):
    '''Replicates common SQLAlchemy attributes from the `source` object to the
    `target` object.'''
    target_manager = manager_of_class(type(target))
    column_attrs = set()
    relationship_attrs = set()
    relationship_columns = set()
    for attr in manager_of_class(type(source)).attributes:
        if attr.key not in target_manager:
            # It's not common attribute
            continue
        target_attr = target_manager[attr.key]
        if isinstance(attr.property, ColumnProperty):
            assert isinstance(target_attr.property, ColumnProperty)
            column_attrs.add(attr)
        elif isinstance(attr.property, RelationshipProperty):
            assert isinstance(target_attr.property, RelationshipProperty)
            relationship_attrs.add(attr)
            if attr.property.direction is MANYTOONE:
                relationship_columns.update(attr.property.local_columns)
    for attr in column_attrs:
        if _column_property_in_registry(attr.property, _excluded):
            continue
        elif (not _column_property_in_registry(attr.property, _included) and
                 all(column in relationship_columns
                     for column in attr.property.columns)):
            continue
        setattr(target, attr.key, getattr(source, attr.key))
    for attr in relationship_attrs:
        target_attr_model = target_manager[attr.key].property.argument
        if not is_relation_replicatable(attr):
            continue
        replicate_relation(source, target, attr, target_manager[attr.key],
                           cache=cache)


def replicate_no_merge(source, model, cache=None):
    '''Replicates the `source` object to `model` class and returns its
    reflection.'''
    # `cache` is used to break circular dependency: we need to replicate
    # attributes before merging target into the session, but replication of
    # some attributes may require target to be in session to avoid infinite
    # loop.
    if source is None:
        return None
    if cache is None:
        cache = {}
    elif source in cache:
        return cache[source]
    db = object_session(source)
    ident = identity_key(instance=source)[1]
    assert ident is not None
    target = db.query(model).get(ident)
    if target is None:
        target = model()
    cache[source] = target
    try:
        replicate_attributes(source, target, cache=cache)
    except _PrimaryKeyIsNull:
        return None
    else:
        return target


def replicate(source, model, cache=None):
    '''Replicates the `source` object to `model` class and returns its
    reflection.'''
    target = replicate_no_merge(source, model, cache=cache)
    if target is not None:
        db = object_session(source)
        target = db.merge(target)
    return target


def replicate_filter(sources, model, cache=None):
    '''Replicates the list of objects to other class and returns their
    reflections'''
    targets = [replicate_no_merge(source, model, cache=cache)
               for source in sources]
    # Some objects may not be available in target DB (not published), so we
    # have to exclude None from the list.
    return [target for target in targets if target is not None]


def reflect_filter(sources, model, cache=None):
    '''Returns the list of reflections of objects in the `source` list to other
    class. Objects that are not found in target table are silently discarded.
    '''
    targets = [reflect(source, model, cache=cache) for source in sources]
    # Some objects may not be available in target DB (not published), so we
    # have to exclude None from the list.
    return [target for target in targets if target is not None]



# -*- coding: utf-8 -*-

from os.path import dirname, abspath, join
import logging
logger = logging.getLogger(__name__)

import jinja2

__all__ = ('TemplateEngine', 'TEMPLATE_DIR')

CURDIR = dirname(abspath(__file__))
TEMPLATE_DIR = join(CURDIR, 'templates')


class TemplateEngine(object):
    '''
    Jinja2 engine adapter.
    '''
    def __init__(self, paths, cache=False, extensions=None):
        '''
        :param paths: list of paths
        :param extensions: list of extensions
        '''
        self.extensions = extensions or []
        self.env = self._make_env(paths)


    def _make_env(self, paths):
        # XXX make an interface method
        return jinja2.Environment(
            loader=jinja2.FileSystemLoader(paths),
            autoescape=True,
            extensions=self.extensions
        )

    def render(self, template_name, **kw):
        'Interface method called from `Template.render`'
        return self.env.get_template(template_name).render(**kw)



# -*- coding: utf-8 -*-

from ..utils import weakproxy
from . import convs


class Widget(object):

    # obsolete parameters from previous versions
    _obsolete = frozenset(['multiple'])

    #: Template to render widget
    template = None
    #: Value of HTML element's *class* attribute
    classname = ''
    #: describes how the widget is rendered.
    #: the following values are supported by default:
    #: 'default': label is rendered in usual place
    #: 'checkbox': label and widget are rendered close to each other
    #: 'full-width': for table-like templates, otherwise should be rendered as default
    #: 'hidden': label is not rendered
    render_type = 'default'
    #: True if widget renders hint itself.
    #: Otherwise parent field should render the hint
    renders_hint = False

    def __init__(self, field=None, **kwargs):
        if self._obsolete & set(kwargs):
            raise TypeError(
                    'Obsolete parameters are used: {}'.format(
                                list(self._obsolete & set(kwargs))))
        self.field = weakproxy(field)
        self._init_kwargs = kwargs
        self.__dict__.update(kwargs)

    @property
    def multiple(self):
        return self.field.multiple

    @property
    def input_name(self):
        return self.field.input_name

    @property
    def id(self):
        return self.field.id

    @property
    def env(self):
        return self.field.env

    def prepare_data(self):
        '''
        Method returning data passed to template.
        Subclasses can override it.
        '''
        value = self.get_raw_value()
        return dict(widget=self,
                    field=self.field,
                    value=value,
                    readonly=not self.field.writable)

    def get_raw_value(self):
        return self.field.raw_value

    def render(self):
        '''
        Renders widget to template
        '''
        data = self.prepare_data()
        if self.field.readable:
            return self.env.template.render(self.template, **data)
        return ''

    def __call__(self, **kwargs):
        '''
        Creates current object's copy with extra constructor arguments passed.
        '''
        kwargs = dict(self._init_kwargs, **kwargs)
        kwargs.setdefault('field', self.field)
        return self.__class__(**kwargs)


class TextInput(Widget):

    template = 'widgets/textinput'
    classname = 'textinput'


class Textarea(Widget):

    template = 'widgets/textarea'


class HiddenInput(Widget):

    render_type = 'hidden'
    template = 'widgets/hiddeninput'


class PasswordInput(Widget):

    template = 'widgets/passwordinput'
    classname = 'textinput'


class Select(Widget):
    '''
    Takes options from :class:`EnumChoice<EnumChoice>` converter,
    looks up if converter allows null and passed this value as template
    :obj:`required` variable.
    '''
    template = 'widgets/select'
    classname = None
    #: HTML select element's select attribute value.
    size = None
    #: Label assigned to None value if field is not required
    null_label = '--------'

    def get_options(self, value):
        options = []

        # XXX ugly
        choice_conv = self.field.conv
        if isinstance(choice_conv, convs.ListOf):
            choice_conv = choice_conv.conv
        assert isinstance(choice_conv, convs.EnumChoice)

        has_null_value = False

        values = value if self.multiple else [value]
        for choice, label in choice_conv.options():
            has_null_value = has_null_value or choice == ''
            options.append(dict(value=choice,
                                title=label,
                                selected=(choice in values)))

        if not self.multiple and not has_null_value and \
                (value == '' or not self.field.conv.required) and \
                self.null_label is not None:
            options.insert(0, {'value': '',
                               'title': self.null_label,
                               'selected': value in (None, '')})
        return options

    def prepare_data(self):
        data = Widget.prepare_data(self)
        return dict(data,
                    options=self.get_options(data['value']),
                    required=('true' if self.field.conv.required else 'false'))


class CheckBoxSelect(Select):

    classname = 'select-checkbox'
    template = 'widgets/select-checkbox'


class CheckBox(Widget):

    render_type = 'checkbox'
    template = 'widgets/checkbox'


class CharDisplay(Widget):

    template = 'widgets/span'
    classname = 'chardisplay'
    #: If is True, value is escaped while rendering.
    #: Passed to template as :obj:`should_escape` variable.
    escape = True
    #: Function converting the value to string.
    getter = staticmethod(lambda v: v)

    def prepare_data(self):
        data = Widget.prepare_data(self)
        return dict(data,
                    value=self.getter(data['value']),
                    should_escape=self.escape)


class AggregateWidget(Widget):

    def get_raw_value(self):
        return None


class FieldListWidget(AggregateWidget):

    allow_create = True
    allow_delete = True

    template = 'widgets/fieldlist'

    def render_template_field(self):
        # used in iktomi.cms: templates/widgets/fieldlist.html
        field = self.field.field(name='%'+self.field.input_name+'-index%')
        # XXX looks like a HACK
        field.set_raw_value(self.field.form.raw_data,
                            field.from_python(field.get_initial()))
        return field.widget.render()


class FieldSetWidget(AggregateWidget):

    template = 'widgets/fieldset'


class FieldBlockWidget(FieldSetWidget):

    render_type = 'full-width'


class FileInput(Widget):

    template = 'widgets/file'




import os, logging
import six
from PIL import Image
from sqlalchemy.orm.session import object_session
from sqlalchemy.orm.util import identity_key
from iktomi.unstable.utils.image_resizers import ResizeFit
from iktomi.utils import cached_property
from iktomi.db.files import TransientFile, PersistentFile
from iktomi.db.sqla.files import FileEventHandlers, FileProperty

logger = logging.getLogger(__name__)


class ImageFile(PersistentFile):

    def _get_properties(self, properties=['width', 'height']):
        if 'width' in properties or 'height' in properties:
            image = Image.open(self.path)
            self.width, self.height = image.size

    @cached_property
    def width(self):
        self._get_properties(['width'])
        return self.width

    @cached_property
    def height(self):
        self._get_properties(['height'])
        return self.height


class ImageEventHandlers(FileEventHandlers):

    def _2persistent(self, target, transient):
        # XXX move this method to file_manager

        # XXX Do this check or not?
        image = Image.open(transient.path)
        assert image.format in Image.SAVE and image.format != 'bmp',\
                'Unsupported image format'

        if self.prop.image_sizes:
            session = object_session(target)
            persistent_name = getattr(target, self.prop.attribute_name)
            if isinstance(persistent_name, six.binary_type):
                persistent_name = persistent_name.decode('utf-8')
            pn, ext = os.path.splitext(persistent_name)

            image_crop = self.prop.resize(image, self.prop.image_sizes)

            if self.prop.force_rgb and image_crop.mode not in ['RGB', 'RGBA']:
                image_crop = image_crop.convert('RGB')
                if ext == '.gif':
                    image_crop.format = 'jpeg'
                    ext = '.jpeg'

            if self.prop.enhancements:
                for enhance, factor in self.prop.enhancements:
                    image_crop = enhance(image_crop).enhance(factor)

            if self.prop.filter:
                image_crop = image_crop.filter(self.prop.filter)

            if not ext:
                # set extension if it is not set
                ext = '.' + image.format.lower()

            if pn + ext != persistent_name:
                persistent_name = pn + ext
                # XXX hack?
                setattr(target, self.prop.attribute_name, persistent_name.encode('utf-8'))

            image_attr = getattr(target.__class__, self.prop.key)
            file_manager = persistent = session.find_file_manager(image_attr)
            persistent = file_manager.get_persistent(persistent_name,
                                                     self.prop.persistent_cls)

            transient = session.find_file_manager(image_attr).new_transient(ext)
            kw = dict(quality=self.prop.quality)
            if self.prop.optimize:
                kw = dict(kw, optimize=True)
            image_crop.save(transient.path, **kw)
            session.find_file_manager(image_attr).store(transient, persistent)
            return persistent
        else:
            # Attention! This method can accept PersistentFile.
            # In this case one shold NEVER been deleted or rewritten.
            assert isinstance(transient, TransientFile), repr(transient)
            return FileEventHandlers._2persistent(self, target, transient)

    def before_update(self, mapper, connection, target):
        FileEventHandlers.before_update(self, mapper, connection, target)
        self._fill_img(mapper, connection, target)

    def before_insert(self, mapper, connection, target):
        FileEventHandlers.before_insert(self, mapper, connection, target)
        self._fill_img(mapper, connection, target)

    def _fill_img(self, mapper, connection, target):
        if self.prop.fill_from:
            # XXX Looks hacky
            value = getattr(target, self.prop.key)
            if value is None:
                base = getattr(target, self.prop.fill_from)
                if base is None:
                    return
                if not os.path.isfile(base.path):
                    logger.warn('Original file is absent %s %s %s',
                                identity_key(instance=target),
                                self.prop.fill_from,
                                base.path)
                    return

                ext = os.path.splitext(base.name)[1]
                session = object_session(target)
                image_attr = getattr(target.__class__, self.prop.key)
                file_manager = session.find_file_manager(image_attr)
                if not getattr(file_manager, "transient_root", False):
                    return
                name = file_manager.new_file_name(
                    self.prop.name_template,
                    target,
                    ext,
                    '',
                )
                setattr(target, self.prop.attribute_name, name)

                persistent = self._2persistent(target, base)
                setattr(target, self.prop.key, persistent)


class ImageProperty(FileProperty):

    event_cls = ImageEventHandlers

    def _set_options(self, options):
        # XXX rename image_sizes?
        options = dict(options)
        self.image_sizes = options.pop('image_sizes', None)

        self.resize = options.pop('resize', None) or ResizeFit()
        # XXX implement
        self.fill_from = options.pop('fill_from', None)
        self.filter = options.pop('filter', None)
        self.enhancements = options.pop('enhancements', [])
        self.force_rgb = self.enhancements or \
                         self.filter or \
                         options.pop('force_rgb', True)
        self.quality = options.pop('quality', 85)
        self.optimize = options.pop('optimize', False)

        assert self.fill_from is None or self.image_sizes is not None

        options.setdefault('persistent_cls', ImageFile)
        FileProperty._set_options(self, options)




# -*- coding: utf-8 -*-

__all__ = ['Application', 'AppEnvironment']

import logging
import re
from iktomi.utils.storage import VersionedStorage, StorageFrame, storage_property
from webob.exc import HTTPException, HTTPInternalServerError, \
                      HTTPNotFound
from webob import Request
from .route_state import RouteState
from .reverse import Reverse

logger = logging.getLogger(__name__)

ip_number = '(\d|[1-9]\d|1\d{2}|2[0-4]\d|25[0-5])'
dns_letter = '[a-z\d]([a-z\d\-]*[a-z\d])?'
port = '(?::\d+)'
HOSTNAME_REGEX = re.compile("^({letter}\.)*{letter}{port}?$".format(letter=dns_letter,
                                                                         port=port), re.I)
IP_REGEX = re.compile("^({number}\.){times}{number}{port}?$".format(number=ip_number,
                                                                    times='{3}',
                                                                    port=port), re.I)

def is_host_valid(host):
    is_hostname = re.match(HOSTNAME_REGEX, host)
    is_ip = re.match(IP_REGEX, host)
    digit_top_domain = re.search("\.\d+$".format(ip_number), host)
    return (is_ip or is_hostname and not digit_top_domain)


class AppEnvironment(StorageFrame):
    '''
    Base class for `env` storage frame class.

    Can be subclassed to add extra functionality::

        class FrontEnvironment(AppEnvironment):
            cfg = cfg
            cache = memcache_client

            @cached_property
            def db(self):
                return db_maker()
    '''

    def __init__(self, request=None, root=None, _parent_storage=None, **kwargs):
        StorageFrame.__init__(self, _parent_storage=_parent_storage, **kwargs)
        self.request = request
        if request:
            self.root = root.bind_to_env(self._root_storage)
            self._route_state = RouteState(request)
        else:
            self.root = root

    def gettext(self, message):
        return message

    def ngettext(self, single, plural, count):
        return single if count == 1 else plural

    @storage_property
    def current_location(self):
        ns = getattr(self, 'namespace', '')
        url_name = getattr(self, 'current_url_name', '')
        return '.'.join(filter(None, (ns, url_name)))

    @classmethod
    def create(cls, *args, **kwargs):
        return VersionedStorage(cls, *args, **kwargs)

    # override this method with configured HelpLoader.get_help to be able to use
    # help messages in admin
    def get_help(*args, **kwargs):
        return ''

class Application(object):
    '''
    WSGI application made from `iktomi.web.WebHandler' instance::

        wsgi_app = Application(app, env_class=FrontEnvironment)
    '''

    env_class = AppEnvironment

    def __init__(self, handler, env_class=None):
        self.handler = handler
        if env_class is not None:
            self.env_class = env_class
        self.root = Reverse.from_handler(handler)

    def handle_error(self, env):
        '''
        Unhandled exception handler.
        You can put any logging, error warning, etc here.'''
        logger.exception('Exception for %s %s :',
                         env.request.method, env.request.url)

    def handle(self, env, data):
        '''
        Calls application and handles following cases:
            * catches `webob.HTTPException` errors.
            * catches unhandled exceptions, calls `handle_error` method
              and returns 500.
            * returns 404 if the app has returned None`.
        '''
        try:
            response = self.handler(env, data)
            if response is None:
                logger.debug('Application returned None '
                             'instead of Response object')
                response = HTTPNotFound()
        except HTTPException as e:
            response = e
        except Exception as e:
            self.handle_error(env)
            response = HTTPInternalServerError()
        return response

    def __call__(self, environ, start_response):
        '''
        WSGI interface method.
        Creates webob and iktomi wrappers and calls `handle` method.
        '''
        # validating Host header to prevent problems with url parsing
        if not is_host_valid(environ['HTTP_HOST']):
            logger.warning('Unusual header "Host: {}", return HTTPNotFound'\
                           .format(environ['HTTP_HOST']))
            return HTTPNotFound()(environ, start_response)
        request = Request(environ, charset='utf-8')
        env = VersionedStorage(self.env_class, request=request, root=self.root)
        data = VersionedStorage()
        response = self.handle(env, data)
        try:
            result = response(environ, start_response)
        except Exception:
            self.handle_error(env)
            result = HTTPInternalServerError()(environ, start_response)
        return result



import warnings, functools

def deprecated(comment=None):
    '''
    This is a decorator which can be used to mark functions
    as deprecated. It will result in a warning being emitted
    when the function is used. Usage::

        @deprecated()
        def foo():
            pass

    or::

        @deprecated('Use bar() instead.')
        def foo():
            pass
    '''
    def deco(func):
        @functools.wraps(func)
        def new_func(*args, **kwargs):
            message = "Call to deprecated function {}.".format(func.__name__)
            if comment is not None:
                message += ' ' + comment
            warnings.warn(message, category=DeprecationWarning, stacklevel=2)
            return func(*args, **kwargs)
        return new_func
    return deco



# -*- coding: utf-8 -*-
import logging
from iktomi.forms import convs, widgets, Field, FieldSet, FileField
from iktomi.utils import cached_property

logger = logging.getLogger(__name__)


class FileFieldSetConv(convs.Converter):

    error_inner = 'Error uploading file'
    error_lost = 'Transient file has been lost'
    error_hacking = 'Transient file name is not provided or is incorrect'

    def from_python(self, value):
        is_transient = value and value.mode == 'transient'
        return {'transient_name': value.name if is_transient else None,
                #'original_name': value and value.original_name,
                'original_name': value.name if is_transient else None, # XXX
                'file': value,
                'mode': value.mode if value is not None else 'empty'}

    @cached_property
    def file_manager(self):
        if hasattr(self.env, 'db') and hasattr(self.env.db, 'find_file_manager'):
            # XXX ducktyping hack?
            return self.env.db.find_file_manager(self.field.form.model)
        return self.env.file_manager

    def _to_python(self, file=None, mode=None,
                   transient_name=None, original_name=None):

        file_manager = self.file_manager

        if not self._is_empty(file):
            file = file_manager.create_transient(file.file, file.filename)
        else:
            file = None

        if mode == 'empty':
            if not file and self.required:
                raise convs.ValidationError(self.error_required)

        elif mode == 'transient':
            if not original_name:
                logger.warning('Missing original_name for FileField')
            if not transient_name:
                logger.warning('Missing transient_name for FileField '
                               'in mode "transient"')
                raise convs.ValidationError(self.error_hacking)

            try:
                transient_file = file_manager.get_transient(transient_name)
            except OSError:
                raise convs.ValidationError(self.error_lost)

            if file:
                file_manager.delete(transient_file)
            else:
                file = transient_file

        elif mode == 'existing':
            if not file:
                return self._existing_value

        return file

    def to_python(self, value):
        value = self._to_python(**value)
        #XXX Hack
        self.field.set_raw_value(self.field.form.raw_data, 
                                 self.from_python(value))
        return value


def check_file_path(conv, value):
    if value and ('/' in value or '\\' in value or value[0] in '.~'):
        logger.warning('Hacking attempt: submitted temp_name '\
                       'for FileField contains "/"')
        raise convs.ValidationError('Invalid filename')
    return value


class FileFieldSet(FieldSet):
    '''FieldSet aggregating fields required for file upload handling::
    '''

    fields = [
            FileField('file',
                  conv=convs.SimpleFile(),
                  widget=widgets.FileInput()),
            Field('mode',
                  conv=convs.EnumChoice(choices=[('existing', ''),
                                                 ('transient', ''),
                                                 ('empty', '')],
                                        required=True),
                  widget=widgets.HiddenInput),
            Field('transient_name',
                  conv=convs.Char(check_file_path, required=False),
                  widget=widgets.HiddenInput),
            Field('original_name',
                  conv=convs.Char(check_file_path, required=False),
                  widget=widgets.HiddenInput),
        ]

    conv = FileFieldSetConv

    def __init__(self, name, **kwargs):
        kwargs.setdefault('fields', self.fields)
        kwargs.setdefault('conv', self.conv)
        FieldSet.__init__(self, name, **kwargs)

    def get_initial(self):
        # Redefine because FieldSet.get_initial returns dict by default,
        # but python value of FileFieldSet is either None, either BaseFile
        # instance.
        return None




# -*- coding: utf-8 -*-

from . import cached_property
import math
import six
import itertools
from six.moves import zip
if six.PY3:# pragma: no cover
    range = lambda *x: list(six.moves.range(*x)) # XXX
from ..web.reverse import URL


def full_page_range(pages_count, page):
    return range(1, pages_count+1)


class FancyPageRange(object):
    '''Insures there are edge pages are shown at each edge and surround pages
    around current page.'''

    def __init__(self, edge=3, surround=5):
        self.edge = edge
        self.surround = surround

    def __call__(self, pages_count, page):
        assert pages_count>0
        assert page>0
        ranges = []
        # left edge
        left_end = min(self.edge, pages_count)
        ranges.append([1, left_end+1])
        # around current page
        surround_start = min(max(1, page-self.surround), pages_count+1)
        surround_end = min(page+self.surround, pages_count)
        if surround_end>=surround_start:
            ranges.append([surround_start, surround_end+1])
        # right edge
        right_start = min(max(1, pages_count-self.edge+1), pages_count+1)
        if pages_count>=right_start:
            ranges.append([right_start, pages_count+1])
        # merge ranges
        ranges.sort()
        pages = range(*ranges[0])
        for current in ranges[1:]:
            last = pages[-1]
            if current[1] <= last:
                # this range is inside previous
                continue
            if current[0] <= last+2:
                # extends previous
                pages += range(last+1, current[1])
            else:
                # comes after with space
                pages += [None] + range(*current)
        return pages


class ChunkedPageRange(object):

    '''Splits pages range into chunks.'''

    def __init__(self, size=10):
        self.paginator_chunk_size = self.size = size

    def __call__(self, pages_count, page):
        assert pages_count>0
        assert page>0
        chunk = (page-1)//self.size
        start = chunk*self.size + 1
        end = min((chunk+1)*self.size, pages_count)
        return range(start, end+1)

    # Everything starting from 'paginator_' is exported to Paginator object.

    def paginator_prev_chunk(paginator):
        chunk = (paginator.page-1)//paginator.chunk_size
        if chunk:
            chunk_size = paginator.chunk_size
            return paginator._page_url_pair((chunk-1)*chunk_size + chunk_size)
        else:
            return paginator._page_url_pair()
    # Due to nature of cached_property we have to explicitly provide name.
    paginator_prev_chunk = cached_property(paginator_prev_chunk, 'prev_chunk')

    def paginator_next_chunk(paginator):
        chunk = (paginator.page-1)//paginator.chunk_size
        page = (chunk+1)*paginator.chunk_size + 1
        if page>paginator.pages_count:
            return paginator._page_url_pair()
        else:
            return paginator._page_url_pair(page)
    # Due to nature of cached_property we have to explicitly provide name.
    paginator_next_chunk = cached_property(paginator_next_chunk, 'next_chunk')


class _PageURL(tuple):

    def __new__(cls, page=None, url=None):
        return tuple.__new__(cls, (page, url))

    @property
    def page(self):
        return self[0]

    @property
    def url(self):
        return self[1]

    def __nonzero__(self):
        return self.page is not None

    __bool__ = __nonzero__


class Paginator(object):
    '''
    Paginator on top of `webob.Request`.
    '''

    #: limit of items on the page
    limit = 0
    #: total count of items
    count = 0
    #: name of GET argument
    page_param = 'page'
    #: show host in URL or not
    show_host = False
    #: items on the current page
    items = ()
    #: Callable returning the list of pages
    #: to show in paginator.
    impl = staticmethod(full_page_range)
    #: The limit of items allowed on the last page. 
    # I.e. if count=23 and orphans=3 with 10 items per page,
    # there will be 2 pages with 10 and 13 items.
    orphans = 0

    def __init__(self, request, **kwargs):
        self.request = request
        self.__dict__.update(kwargs)
        # Calculate page here in case we return 404 for invalid page
        self.page

    def __nonzero__(self):
        '''Should paginator be shown?'''
        # We have to check current page too in case the list of items just
        # shrunk to fit into one page.
        return bool(self.limit) and (self.pages_count>1 or self.page>1)

    __bool__ = __nonzero__

    def invalid_page(self):
        '''This method is called when invalid (not positive int)
        page is passed in request.
        Use "pass" to ignore. Other options are raising exceptions for HTTP
        Not Found or redirects.'''
        pass

    @cached_property
    def page(self):
        '''Current page.'''
        page = self.request.GET.get(self.page_param)
        if not page:
            return 1
        try:
            page = int(page)
        except ValueError:
            self.invalid_page()
            return 1
        if page<1:
            self.invalid_page()
            return 1
        return page

    @cached_property
    def url(self):
        '''Current or base URL. Can be redefined via keyword argument on
        initialization.

        Returns `iktomi.web.URL object.
        `'''
        return URL.from_url(self.request.url, show_host=self.show_host)

    def page_url(self, page):
        '''
        Returns URL for page, page is included as query parameter.

        Can be redefined by keyword argument
        '''
        if page is not None and page != 1:
            return self.url.qs_set(**{self.page_param: page})
        elif page is not None:
            return self.url.qs_delete('page')

    def _page_url_pair(self, page=None):
        return _PageURL(page, self.page_url(page) if page is not None else None)

    @cached_property
    def pages_count(self):
        '''Number of pages.'''
        if not self.limit or self.count<self.limit:
            return 1
        if self.count % self.limit <= self.orphans:
            return self.count // self.limit
        return int(math.ceil(float(self.count)/self.limit))

    def slice(self, items):
        '''Slice the sequence of all items to obtain them for current page.'''
        if self.limit:
            if self.page>self.pages_count:
                return []
            if self.page == self.pages_count:
                return items[self.limit*(self.page-1):]
            return items[self.limit*(self.page-1):self.limit*self.page]
        else:
            return items[:]

    def enumerate(self):
        skipped = (self.page-1)*self.limit
        return zip(itertools.count(skipped+1), self.items)

    @cached_property
    def prev(self):
        if self.page>self.pages_count:
            # if we are on non-existing page which is higher
            # than current page, return last page, so we can navigate
            # to it
            return self.last
        elif self.page>1:
            return self._page_url_pair(self.page-1)
        else:
            return self._page_url_pair()

    @cached_property
    def next(self):
        if self.page<self.pages_count:
            return self._page_url_pair(self.page+1)
        else:
            return self._page_url_pair()

    @cached_property
    def first(self):
        return self._page_url_pair(1)

    @cached_property
    def last(self):
        return self._page_url_pair(self.pages_count)

    @cached_property
    def pages(self):
        pages = self.impl(self.pages_count, self.page)
        return [self._page_url_pair(page) for page in pages]

    def __getattr__(self, requested_name):
        name = 'paginator_'+requested_name
        impl_type = type(self.impl)
        # Look for exported descriptors defined in impl first.
        # We can't lookup in __dict__ since this doesn't work for inheritted
        # attributes.
        if name not in dir(impl_type):
            return getattr(self.impl, name)
        prop = getattr(impl_type, name)
        # The following will work only if descriptor returns self when
        # accessed in class.
        if hasattr(prop, '__get__'):
            return prop.__get__(self, type(self))
        return prop # pragma: no cover


class ModelPaginator(Paginator):

    '''
    Paginator for sqlalchemy query.
    '''

    def __init__(self, request, query, **kwargs):
        self._query = query
        Paginator.__init__(self, request, **kwargs)

    @cached_property
    def count(self):
        return self._query.count()

    @cached_property
    def items(self):
        '''Items on the current page. Filled in automatically based on query
        and currrent page'''
        return self.slice(self._query)

    def __getitem__(self, key):
        return self.items[key]

    def __iter__(self):
        return iter(self.items)

    def __len__(self):
        return len(self.items)
